"use strict";
/*
 * Copyright (c) 2017, Salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
Object.defineProperty(exports, "__esModule", { value: true });
const path = require('path');
const MetadataRegistry = require(path.join(__dirname, 'metadataRegistry'));
const glob = require('glob');
const PathUtil = require("./sourcePathUtil");
/**
 * This class contains helper functions to calculate Bundle type paths (AuraDefinitionBundle, LightningComponentBundle)
 */
class BundlePathHelper {
    static getPathFromBundleTypeFileProperties(fullName, defaultSourceDir, defaultDirName, bundleFileTypeProperties) {
        const bundleName = fullName.split(path.sep)[0];
        const pathToDefaultTypeDirectory = path.join(defaultSourceDir, defaultDirName);
        const fileName = BundlePathHelper.getMetadataFileNameFromBundleFileProperties(fullName, bundleFileTypeProperties);
        return path.join(pathToDefaultTypeDirectory, bundleName, fileName);
    }
    ;
    static getMetadataFileNameFromBundleFileProperties(fullName, bundleFileTypeProperties) {
        const bundleName = fullName.split(path.sep)[0];
        const matchingBundle = bundleFileTypeProperties.find(fileProperty => fileProperty.fullName.split(path.sep)[0] === bundleName);
        return `${path.basename(matchingBundle.fileName)}${MetadataRegistry.getMetadataFileExt()}`;
    }
    static findMetadataFilePathInBundleDir(filePath, bundleTypeDefaultDirName) {
        if (filePath.endsWith(MetadataRegistry.getMetadataFileExt())) {
            return filePath;
        }
        else {
            // Each bundle has one metadata file path that ends with its definition suffix + '-meta.xml'
            // and is located directly in the bundle directory (not a subfolder)
            // This finds and scans the bundle directory to find the existing aggregate metadata file path
            let bundlePath;
            let pathName = path.dirname(filePath);
            let folderName = path.basename(pathName);
            // LightningComponentBundle folder structure can be multi-level; Scan for the direct child folder
            // which is the bundle path
            while (!(folderName === bundleTypeDefaultDirName)) {
                bundlePath = pathName;
                pathName = path.dirname(pathName);
                folderName = path.basename(pathName);
            }
            const matchingFilePaths = glob.sync(path.join(bundlePath, `*${MetadataRegistry.getMetadataFileExt()}`));
            if (matchingFilePaths.length > 0) {
                // glob returns paths using the forward slash only, which breaks in Windows
                return PathUtil.replaceForwardSlashes(matchingFilePaths[0]);
            }
            // In the case that a bundle has been deleted from the workspace, a search for its definition file
            // will produce no results. In this case, return the filePath being processed.
            return filePath;
        }
    }
    static buildFullNameFromFilePath(filePath, defaultDirectory) {
        const filePathWithoutMetadataExt = PathUtil.removeMetadataFileExtFrom(filePath);
        let pathName = path.dirname(filePathWithoutMetadataExt);
        let bundleName = path.basename(pathName);
        const fileName = path.basename(filePathWithoutMetadataExt);
        let fullName = path.join(bundleName, fileName);
        // Walk back up directory structure to build fullName to include subfolder hierarchy within the fullName
        while (!(path.basename(path.dirname(pathName)) === defaultDirectory)) {
            pathName = path.dirname(pathName);
            bundleName = path.basename(pathName);
            fullName = path.join(bundleName, fullName);
        }
        return fullName;
    }
    static scanFilePathForAggregateFullName(filePath, defaultDirectory) {
        let aggregateFullName;
        let pathName = path.dirname(filePath);
        let folderName = path.basename(pathName);
        // Some bundle folder structure can be multi-level; Scan for the direct child folder
        // which is the bundle name
        while (!(folderName === defaultDirectory)) {
            aggregateFullName = folderName;
            pathName = path.dirname(pathName);
            folderName = path.basename(pathName);
        }
        return aggregateFullName;
    }
    static getAllNestedBundleContentPaths(bundleDirPath, forceIgnore) {
        const bundlePaths = glob.sync(path.join(bundleDirPath, '**'), { nodir: true });
        return Promise.resolve(bundlePaths.map(bundlePath => PathUtil.replaceForwardSlashes(bundlePath))
            .filter(bundlePath => forceIgnore.accepts(bundlePath) &&
            !bundlePath.endsWith(MetadataRegistry.getMetadataFileExt())));
    }
    static getExtendedBundlePath(fileName, bundleName) {
        // Is there a nested directory or directories; if so this takes the full file name and
        // returns the directory path between the file and bundeName
        let extendedPath;
        if (fileName && fileName.includes(bundleName)) {
            let nestedDir = path.dirname(fileName);
            let nestedDirName = path.basename(nestedDir);
            while (nestedDirName !== bundleName) {
                if (extendedPath) {
                    extendedPath = path.join(nestedDirName, extendedPath);
                }
                else {
                    extendedPath = nestedDirName;
                }
                nestedDir = path.dirname(nestedDir);
                nestedDirName = path.basename(nestedDir);
            }
        }
        return extendedPath;
    }
}
exports.BundlePathHelper = BundlePathHelper;

//# sourceMappingURL=bundlePathHelper.js.map
