/*
 * Copyright (c) 2016, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
/* --------------------------------------------------------------------------------------------------------------------
 * WARNING: This file has been deprecated and should now be considered locked against further changes.  Its contents
 * have been partially or wholely superceded by functionality included in the @salesforce/core npm package, and exists
 * now to service prior uses in this repository only until they can be ported to use the new @salesforce/core library.
 *
 * If you need or want help deciding where to add new functionality or how to migrate to the new library, please
 * contact the CLI team at alm-cli@salesforce.com.
 * ----------------------------------------------------------------------------------------------------------------- */
'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
// Node
const path = require('path');
const _ = require('lodash');
// Local
const ConfigFile = require(path.join(__dirname, 'ConfigFile'));
const almError = require(path.join(__dirname, '..', 'almError'));
const consts = require(path.join(__dirname, '..', 'constants'));
const messages = require(path.join(__dirname, '..', 'messages'))();
const srcDevUtil = require(path.join(__dirname, '..', 'srcDevUtil'));
const Crypto = require(path.join(__dirname, '..', 'crypto'));
const SFDX_CONFIG_FILE_NAME = 'sfdx-config.json';
const OrgDefaults = {
    DEVHUB: consts.DEFAULT_DEV_HUB_USERNAME,
    USERNAME: consts.DEFAULT_USERNAME,
    list() {
        return [OrgDefaults.DEVHUB, OrgDefaults.USERNAME, OrgDefaults.APIVERSION];
    }
};
// In the future consider keying input validation off of command type validation.
const ALLOWED_PROPERTIES = [
    {
        key: 'instanceUrl',
        input: {
            // If a value is provided validate it otherwise no value is unset.
            validator: value => _.isNil(value) || srcDevUtil.isSalesforceDomain(value),
            failedMessage: messages.getMessage('invalidInstanceUrl', null, 'sfdxConfig')
        }
    },
    {
        key: 'apiVersion',
        hidden: true,
        input: {
            // If a value is provided validate it otherwise no value is unset.
            validator: (value) => _.isNil(value) || /[1-9]\d\.0/.test(value),
            failedMessage: messages.getMessage('invalidApiVersion', null, 'sfdxConfig')
        }
    },
    { key: consts.DEFAULT_DEV_HUB_USERNAME },
    { key: consts.DEFAULT_USERNAME },
    {
        key: 'restDeploy',
        hidden: true,
        input: {
            validator: (value) => value.toString() === 'true' || value.toString() === 'false',
            failedMessage: messages.getMessage('invalidBooleanConfigValue')
        }
    }, {
        key: 'isvDebuggerSid',
        encrypted: true
    }, {
        key: 'isvDebuggerUrl'
    }
];
const propertyConfigMap = _.keyBy(ALLOWED_PROPERTIES, 'key');
class SfdxConfig extends ConfigFile {
    constructor(isGlobal) {
        super(SFDX_CONFIG_FILE_NAME, isGlobal, true);
    }
    initCrypto() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.crypto) {
                this.crypto = new Crypto();
                yield this.crypto.init();
            }
        });
    }
    clearCrypto() {
        if (this.crypto) {
            this.crypto.close();
            delete this.crypto;
        }
    }
    cryptProperties(encrypt) {
        return __awaiter(this, void 0, void 0, function* () {
            const hasEncryptedProperties = _.some(this.contents, (val, key) => !!this.getPropertyConfig(key).encrypted);
            // Only initialize crypto if there are properties that are encrypted.
            if (hasEncryptedProperties) {
                yield this.initCrypto();
                _.each(this.contents, (value, key) => {
                    if (this.getPropertyConfig(key).encrypted) {
                        this.contents[key] = encrypt ?
                            this.crypto.encrypt(value) : this.crypto.decrypt(value);
                    }
                });
            }
        });
    }
    read() {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield _super("read").call(this);
                yield this.cryptProperties(false);
            }
            finally {
                this.clearCrypto();
            }
            return Promise.resolve(this.contents);
        });
    }
    write() {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            try {
                // Encrypt the properties
                yield this.cryptProperties(true);
                yield _super("write").call(this);
                // Decrypt after write so they can still be used
                yield this.cryptProperties(false);
            }
            finally {
                this.clearCrypto();
            }
        });
    }
    /**
     * Sync version of read.
     * @deprecated This should only be used by legacy config.
     */
    readSync() {
        try {
            this.contents = srcDevUtil.readJSONSync(this.path, false);
        }
        catch (err) {
            return this.checkEnoent(err);
        }
        return this.contents;
    }
    getPropertyConfig(propertyName) {
        const property = propertyConfigMap[propertyName];
        if (!property) {
            throw almError({ keyName: 'UnknownConfigKey', bundle: 'sfdxConfig' }, [propertyName]);
        }
        return property;
    }
    setPropertyValue(propertyName, value) {
        const property = this.getPropertyConfig(propertyName);
        if (_.isNil(value)) {
            delete this.contents[propertyName];
        }
        else if (property.input) {
            if (property.input && property.input.validator(value)) {
                this.contents[property.key] = value;
            }
            else {
                throw almError({ keyName: 'invalidConfigValue', bundle: 'sfdxConfig' }, [property.input.failedMessage]);
            }
        }
        else {
            this.contents[property.key] = value;
        }
    }
    static getAllowedProperties() {
        return ALLOWED_PROPERTIES;
    }
    static set(isGlobal, property, value) {
        return __awaiter(this, void 0, void 0, function* () {
            const config = new SfdxConfig(isGlobal);
            yield config.read();
            config.setPropertyValue(property, value);
            return config.write();
        });
    }
    static clear() {
        return __awaiter(this, void 0, void 0, function* () {
            // Clear the global config
            yield new SfdxConfig(true).clear();
            // Then clear the local config
            yield new SfdxConfig(false).clear();
        });
    }
}
SfdxConfig.OrgDefaults = OrgDefaults;
module.exports = SfdxConfig;

//# sourceMappingURL=SfdxConfig.js.map
