/*
 * Copyright (c) 2016, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
// 3pp
const Promise = require('bluebird');
const isNil = require('lodash/isNil');
// Node
const path = require('path');
// Local
const logger = require(path.join('..', 'lib', 'logApi'));
const { MdRetrieveApi } = require(path.join(__dirname, 'mdapiRetrieveApi'));
const SourceWorkspaceAdapter = require(path.join(__dirname, 'sourceWorkspaceAdapter'));
const srcDevUtil = require(path.join(__dirname, 'srcDevUtil'));
const SourceUtil = require(path.join(__dirname, 'source', 'sourceUtil'));
const ManifestCreateApi = require(path.join(__dirname, 'manifestCreateApi'));
const SourceMetadataMemberRetrieveHelper = require(path.join(__dirname, 'sourceMetadataMemberRetrieveHelper'));
const syncCommandHelper = require(path.join(__dirname, 'syncCommandHelper'));
const messagesApi = require(path.join(__dirname, 'messages'));
const MetadataRegistry = require(path.join(__dirname, 'metadataRegistry'));
const AuraDefinitionBundleMetadataType = require(path.join(__dirname, 'metadataTypeImpl', 'auraDefinitionBundleMetadataType')).AuraDefinitionBundleMetadataType;
const srcStatusApi = require(path.join(__dirname, 'srcStatusApi'));
const pathUtil = require(path.join(__dirname, 'sourcePathUtil'));
class MdapiPullApi {
    constructor(scratchOrg, sourceWorkspaceAdapter) {
        this.scratchOrg = scratchOrg;
        this.force = scratchOrg.force;
        this.logger = logger.child('md-pull');
        this.swa = sourceWorkspaceAdapter || new SourceWorkspaceAdapter(this.scratchOrg, this.force.getConfig().getAppConfig().defaultPackagePath, MetadataRegistry);
        this.maxRevisionFile = this.scratchOrg.getMaxRevision();
        this.smmHelper = new SourceMetadataMemberRetrieveHelper(this.swa.metadataRegistry);
        this.messages = messagesApi(this.force.config.getLocale());
        this.obsoleteNames = [];
    }
    doPull(options) {
        return __awaiter(this, void 0, void 0, function* () {
            // Remove this when pull has been modified to support the new mdapi wait functionality;
            if (isNaN(options.wait)) {
                options.wait = this.force.config.getConfigContent().defaultSrcWaitMinutes;
            }
            yield this._checkForConflicts(options);
            const maxRevision = yield srcDevUtil.getMaxRevision(this.maxRevisionFile.path);
            const pkg = yield this.smmHelper.getRevisionsAsPackage(maxRevision, this.obsoleteNames);
            try {
                // Create a temp directory
                options.retrievetargetdir = yield SourceUtil.createOutputDir('pull');
                // Create a manifest (package.xml).
                const manifestOptions = Object.assign({}, options, { outputdir: options.retrievetargetdir });
                const manifest = yield this._createPackageManifest(manifestOptions, pkg);
                let result;
                if (manifest.empty) {
                    if (this.obsoleteNames.length > 0) {
                        result = { fileProperties: [], success: true, status: 'Succeeded' };
                    }
                }
                else {
                    // Get default metadata retrieve options
                    const retrieveOptions = Object.assign(MdRetrieveApi.getDefaultOptions(), {
                        retrievetargetdir: options.retrievetargetdir,
                        unpackaged: manifest.file,
                        wait: options.wait
                    });
                    // Retrieve the metadata
                    result = yield new MdRetrieveApi(this.scratchOrg).retrieve(retrieveOptions).catch((err) => err.result);
                }
                // Update local metadata source.
                return yield this._postRetrieve(result, options);
            }
            finally {
                // Delete the output dir.
                yield SourceUtil.cleanupOutputDir(options.retrievetargetdir);
            }
        });
    }
    _createPackageManifest(options, pkg) {
        return __awaiter(this, void 0, void 0, function* () {
            if (pkg.isEmpty()) {
                return Promise.resolve({ empty: true });
            }
            if (isNil(options.packageXml) || !options.debug) {
                const configSourceApiVersion = this.force.getConfig().getAppConfig().sourceApiVersion;
                const sourceApiVersion = !isNil(configSourceApiVersion) ? configSourceApiVersion : this.force.getConfig().getApiVersion();
                pkg.setVersion(sourceApiVersion);
                return Promise.resolve(new ManifestCreateApi(this.force).createManifestForMdapiPackage(options, pkg, this.smmHelper.metadataRegistry));
            }
            else {
                return Promise.resolve({ file: options.packageXml });
            }
        });
    }
    static _didRetrieveSucceed(result) {
        return !isNil(result) &&
            result.success && result.status === 'Succeeded' &&
            isNil(result.messages) &&
            !isNil(result.fileProperties) && Array.isArray(result.fileProperties);
    }
    _postRetrieve(result, options) {
        return __awaiter(this, void 0, void 0, function* () {
            let changedSourceElements;
            if (MdapiPullApi._didRetrieveSucceed(result)) {
                changedSourceElements = yield this._syncDownSource(result, options, this.swa);
                yield this._updateMaxRevision();
            }
            return this._processResults(result, changedSourceElements);
        });
    }
    _syncDownSource(result, options, swa) {
        const changedSourceElements = new Map();
        // Each Aura bundle has a definition file that has one of the suffixes: .app, .cmp, .design, .evt, etc.
        // In order to associate each sub-component of an aura bundle (e.g. controller, style, etc.) with
        // its parent aura definition type, we must find its parent's file properties
        // and pass those along to processMdapiFileProperty
        const auraDefinitionFileFileProperties = AuraDefinitionBundleMetadataType.getAuraDefinitionFileFileProperties(result.fileProperties, this.swa.metadataRegistry);
        result.fileProperties.forEach((fileProperty) => {
            if (fileProperty.type === 'Package') {
                return;
            }
            // After retrieving, switch back to path separators (for Windows)
            fileProperty.fullName = pathUtil.replaceForwardSlashes(fileProperty.fullName);
            this.swa.processMdapiFileProperty(changedSourceElements, options.retrievetargetdir, fileProperty, auraDefinitionFileFileProperties);
        });
        this.obsoleteNames.forEach((obsoleteName) => {
            this.swa.handleObsoleteSource(changedSourceElements, obsoleteName.fullName, obsoleteName.type);
        });
        return swa.updateSource(changedSourceElements, this.force, this.scratchOrg, options.manifest, false /** check for duplicates **/, options.unsupportedMimeTypes);
    }
    _updateMaxRevision() {
        return this.smmHelper.getLatestSourceRevisionCount()
            .then((newMaxRev) => Promise.resolve(this.maxRevisionFile.write(parseInt(newMaxRev))));
    }
    _processResults(result, changedSourceElements) {
        if (isNil(result)) {
            return Promise.resolve();
        }
        else if (MdapiPullApi._didRetrieveSucceed(result)) {
            const inboundFiles = [...changedSourceElements.values()]
                .reduce((allChanges, sourceElement) => allChanges.concat(sourceElement.getWorkspaceElements()), [])
                .map(workspaceElement => workspaceElement.toObject());
            return Promise.resolve({ inboundFiles });
        }
        else {
            const retrieveFailed = new Error(syncCommandHelper.getRetrieveFailureMessage(result, this.messages));
            retrieveFailed.name = 'RetrieveFailed';
            return Promise.reject(retrieveFailed);
        }
    }
    _checkForConflicts(options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (options.forceoverwrite) { // do not check for conflicts when pull --forceoverwrite
                return [];
            }
            const statusApi = new srcStatusApi(this.scratchOrg, this.swa);
            return statusApi.doStatus({ local: true, remote: true }) // rely on status so that we centralize the logic
                .then(() => statusApi.getLocalConflicts())
                .then((conflicts) => {
                if (conflicts.length > 0) {
                    const error = new Error('Conflicts found during sync down');
                    error.name = 'SourceConflict';
                    error.sourceConflictElements = conflicts;
                    throw error;
                }
            });
        });
    }
}
module.exports = MdapiPullApi;

//# sourceMappingURL=mdapiPullApi.js.map
