"use strict";
/*
 * Copyright (c) 2016, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
const path = require('path');
const Promise = require('bluebird');
const util = require('util');
const messages = require(path.join(__dirname, 'messages'))();
const almError = require(path.join(__dirname, 'almError'));
const srcDevUtil = require(path.join(__dirname, 'srcDevUtil.js'));
const { parseWaitParam } = require(path.join(__dirname, 'source', 'sourceUtil'));
const MdapiPullApi = require(path.join(__dirname, 'mdapiPullApi'));
const syncCommandHelper = require(path.join(__dirname, 'syncCommandHelper'));
const MetadataRegistry = require(path.join(__dirname, 'metadataRegistry'));
const logger = require(path.join(__dirname, 'logApi'));
const { Messages } = require('@salesforce/core');
Messages.importMessagesDirectory(__dirname);
class MdapiPullCommand {
    constructor() {
        this.logger = logger.child('source:pull');
        this.commonMsgs = Messages.loadMessages('salesforce-alm', 'source');
    }
    validate(context) {
        // Validate the wait param if set and convert to an integer.
        parseWaitParam(context.flags);
        return Promise.resolve(context);
    }
    getPreExecuteMessage({ orgId, username }) {
        return messages.getMessage('pullCommandCliPreExecute', [orgId, username]);
    }
    execute(context) {
        const scratchOrg = context.org;
        const force = scratchOrg.force;
        const options = context.flags || {};
        const rows = [];
        const projectPath = force.config.getProjectPath();
        let mdapiPull;
        return MetadataRegistry.initializeMetadataTypeInfos(scratchOrg)
            .then(() => {
            mdapiPull = new MdapiPullApi(scratchOrg);
            options.unsupportedMimeTypes = []; // for logging unsupported static resource mime types
            return mdapiPull.doPull(options);
        })
            .catch(e => {
            if (e.name === 'SourceConflict') {
                const error = almError('sourceConflictDetected');
                e.sourceConflictElements.forEach(sourceElement => syncCommandHelper.createConflictRows(rows, sourceElement, projectPath));
                error.columns = syncCommandHelper.getColumnMetaInfo(this.commonMsgs);
                error.result = rows;
                this.logger.error(messages.getMessage('pullCommandConflictMsg'));
                throw error;
            }
            else {
                throw e;
            }
        })
            .then((result) => {
            if (!util.isNullOrUndefined(result)) {
                result.inboundFiles.forEach((sourceItem) => syncCommandHelper.createDisplayRows(rows, sourceItem, projectPath, false));
            }
        })
            .then(() => this.logger.styledHeader(this.logger.color.blue(messages.getMessage('pullCommandHumanSuccess'))))
            .then(() => srcDevUtil.logUnsupportedMimeTypeError(options.unsupportedMimeTypes, this.logger, force))
            .then(() => (options.json ? { pulledSource: rows } : rows));
    }
    /**
     * this indicated to index.js this command should produce tabular output.
     * @returns {*[]}
     */
    getColumnData() {
        return syncCommandHelper.getColumnMetaInfo(this.commonMsgs);
    }
}
module.exports = MdapiPullCommand;

//# sourceMappingURL=mdapiPullCommand.js.map
