"use strict";
/*
 * Copyright (c) 2017, Salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
Object.defineProperty(exports, "__esModule", { value: true });
const path = require('path');
const glob = require('glob');
const srcDevUtil = require(path.join(__dirname, '..', 'srcDevUtil'));
const messages = require(path.join(__dirname, '..', 'messages'))();
const MetadataRegistry = require(path.join(__dirname, '..', 'metadataRegistry'));
const defaultMetadataType_1 = require("./defaultMetadataType");
const metadataTypeFactory_1 = require("../metadataTypeFactory");
const PathUtil = require("../sourcePathUtil");
const bundlePathHelper_1 = require("../bundlePathHelper");
class AuraDefinitionBundleMetadataType extends defaultMetadataType_1.DefaultMetadataType {
    constructor(typeDefObj) {
        super(typeDefObj);
    }
    getFullNameFromFilePath(filePath) {
        const bundleName = PathUtil.getParentDirectoryName(filePath);
        const fileName = PathUtil.removeMetadataFileExtFrom(path.basename(filePath));
        return path.join(bundleName, fileName);
    }
    getAggregateFullNameFromFilePath(filePath) {
        return PathUtil.getParentDirectoryName(filePath);
    }
    getAggregateMetadataFilePathFromWorkspacePath(filePath) {
        return bundlePathHelper_1.BundlePathHelper.findMetadataFilePathInBundleDir(filePath, this.typeDefObj.defaultDirectory);
    }
    getDefaultAggregateMetadataPath(fullName, defaultSourceDir, bundleFileProperties) {
        return bundlePathHelper_1.BundlePathHelper.getPathFromBundleTypeFileProperties(fullName, defaultSourceDir, this.typeDefObj.defaultDirectory, bundleFileProperties);
    }
    getAggregateFullNameFromSourceMemberName(sourceMemberName) {
        return sourceMemberName.split(path.sep)[0];
    }
    getAggregateFullNameFromWorkspaceFullName(workspaceFullName) {
        return workspaceFullName.split(path.sep)[0];
    }
    getAggregateFullNameFromFileProperty(fileProperty, namespace) {
        return fileProperty.fullName.split(path.sep)[0];
    }
    getPathToMdapiSourceDir(aggregateFullName, mdDir) {
        return path.join(mdDir, this.typeDefObj.defaultDirectory, aggregateFullName);
    }
    getRetrievedMetadataPath(fileProperty, retrieveRoot, bundleFileProperties) {
        const bundlePath = path.dirname(fileProperty.fileName);
        const fileName = bundlePathHelper_1.BundlePathHelper.getMetadataFileNameFromBundleFileProperties(fileProperty.fullName, bundleFileProperties);
        const retrievedMetadataPath = path.join(retrieveRoot, bundlePath, fileName);
        if (srcDevUtil.pathExistsSync(retrievedMetadataPath)) {
            return retrievedMetadataPath;
        }
        return null;
        // W-3903546 older AuraDefinitionBundles do not require a -meta.xml file
    }
    getOriginContentPathsForSourceConvert(metadataFilePath, workspaceVersion, unsupportedMimeTypes, forceIgnore) {
        const bundleDirPath = path.dirname(metadataFilePath);
        const bundlePaths = glob.sync(path.join(bundleDirPath, '*'));
        return Promise.resolve(bundlePaths.filter((bundlePath) => forceIgnore.accepts(bundlePath) &&
            !bundlePath.endsWith(MetadataRegistry.getMetadataFileExt())));
    }
    sourceMemberFullNameConflictsWithWorkspaceFullName(sourceMemberFullName, workspaceFullName) {
        const aggregateSourceMemberName = this.getAggregateFullNameFromSourceMemberName(sourceMemberFullName);
        const aggregateFullName = this.getAggregateFullNameFromWorkspaceFullName(workspaceFullName);
        return aggregateSourceMemberName === aggregateFullName;
    }
    trackRemoteChangeForSourceMemberName(sourceMemberName) {
        // Whenever a component of an AuraBundle is modified in the scratch org, a SourceMember is created
        // at the bundle level and another is created for the changed component. Ignore the SourceMember created
        // for the bundle and track specific bundle component changes only.
        return sourceMemberName.split(path.sep).length > 1;
    }
    onlyDisplayOneConflictPerAggregate() {
        // Waiting on W-4389673 to support fine-grained push and pull of AuraDefinitionBundle components
        // Until they are supported, we only want to report one conflict entry per bundle
        return true;
    }
    getDisplayPathForLocalConflict(workspaceFilePath) {
        return path.dirname(workspaceFilePath);
    }
    getAggregateFullNameFromComponentFailure(componentFailure) {
        // if it's an aura-related error, the fullame is the name of the file within the bundle without any extension
        // for example for a bundle with fullname=testbundle and an error in the helper the MD API response will contain:
        // "fileName": "metadataPackage_1497651865666/aura/testbundle/testbundleHelper.js",
        // "fullName": "testbundleHelper",
        return this.getAggregateFullNameFromFilePath(componentFailure.fileName);
    }
    getAggregateFullNameFromMdapiPackagePath(mdapiPackagePath) {
        const pathElements = mdapiPackagePath.split(path.sep);
        return pathElements[1];
    }
    validateDeletedContentPath(deletedContentPath, contentPaths, metadataRegistry) {
        // For Aura Bundles, if we are deleting the metadata file, but there are still other aura definitions present, we should throw an error
        const auraTypeDef = metadataRegistry.getLightningDefByFileName(deletedContentPath);
        if (auraTypeDef && auraTypeDef.hasMetadata) {
            const otherAuraFilesExist = contentPaths.some(contentPath => !contentPath.endsWith(auraTypeDef.fileSuffix));
            if (otherAuraFilesExist) {
                const err = new Error();
                err.message = messages.getMessage('MissingContentFile', deletedContentPath);
                err.name = 'Missing Content File';
                throw err;
            }
        }
    }
    getWorkspaceFullNameFromComponentFailure(componentFailure) {
        return this.getFullNameFromFilePath(componentFailure.fileName);
    }
    parseSourceMemberForMetadataRetrieve(sourceMemberName, sourceMemberType, isNameObsolete) {
        sourceMemberName = PathUtil.replaceForwardSlashes(sourceMemberName);
        return { fullName: sourceMemberName, type: sourceMemberType, isNameObsolete };
    }
    /**
     * Returns the fileProperty object for the Aura definition metadata file corresponding to the given filePropertyFileName
     * @param retrieveRoot
     * @param filePropertyFileName
     * @returns {any}
     */
    static getCorrespondingAuraDefinitionFileProperty(retrieveRoot, filePropertyFileName, auraMetadataName, metadataRegistry) {
        const bundleDirPath = path.join(retrieveRoot, path.dirname(filePropertyFileName));
        const bundlePaths = glob.sync(path.join(bundleDirPath, '*'));
        const bundleDefinitionPath = bundlePaths.find(bundlePath => AuraDefinitionBundleMetadataType.isDefinitionFile(bundlePath, metadataRegistry));
        const auraDefinitionFileProperty = {
            type: auraMetadataName,
            fileName: path.relative(retrieveRoot, bundleDefinitionPath),
            fullName: path.basename(bundleDefinitionPath, path.extname(bundleDefinitionPath))
        };
        return auraDefinitionFileProperty;
    }
    static getAuraDefinitionFileFileProperties(fileProperties, metadataRegistry) {
        return fileProperties.filter(fileProperty => {
            const metadataType = metadataTypeFactory_1.MetadataTypeFactory.getMetadataTypeFromMetadataName(fileProperty.type, metadataRegistry);
            if (metadataType instanceof AuraDefinitionBundleMetadataType) {
                return AuraDefinitionBundleMetadataType.isDefinitionFile(fileProperty.fileName, metadataRegistry);
            }
            return false;
        }).map((fileProperty) => {
            fileProperty.fullName = PathUtil.replaceForwardSlashes(fileProperty.fullName);
            fileProperty.fileName = PathUtil.replaceForwardSlashes(fileProperty.fileName);
            return fileProperty;
        });
    }
    static isDefinitionFile(filePath, metadataRegistry) {
        const lightningDefType = metadataRegistry.getLightningDefByFileName(filePath);
        if (lightningDefType) {
            const isDefinitionFile = lightningDefType.hasMetadata;
            return isDefinitionFile;
        }
        return false;
    }
    shouldDeleteWorkspaceAggregate(metadataType) {
        // Handle deletes of AuraDefinitionBundles at the subcomponent level because
        // SourceMembers are created for each subcomponent
        return false;
    }
    handleSlashesForSourceMemberName(sourceMemberFullName) {
        return PathUtil.replaceForwardSlashes(sourceMemberFullName);
    }
    sourceMemberFullNameCorrespondsWithWorkspaceFullName(sourceMemberFullName, workspaceFullName) {
        return sourceMemberFullName === workspaceFullName;
    }
}
exports.AuraDefinitionBundleMetadataType = AuraDefinitionBundleMetadataType;

//# sourceMappingURL=auraDefinitionBundleMetadataType.js.map
