"use strict";
/*
 * Copyright (c) 2017, Salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
Object.defineProperty(exports, "__esModule", { value: true });
const path = require('path');
const glob = require('glob');
const messages = require(path.join(__dirname, '..', 'messages'))();
const MetadataRegistry = require(path.join(__dirname, '..', 'metadataRegistry'));
const srcDevUtil = require(path.join(__dirname, '..', 'srcDevUtil'));
const defaultMetadataType_1 = require("./defaultMetadataType");
const bundlePathHelper_1 = require("../bundlePathHelper");
const PathUtil = require("../sourcePathUtil");
class WaveTemplateBundleMetadataType extends defaultMetadataType_1.DefaultMetadataType {
    constructor(typeDefObj) {
        super(typeDefObj);
    }
    getFullNameFromFilePath(filePath) {
        return bundlePathHelper_1.BundlePathHelper.buildFullNameFromFilePath(filePath, this.typeDefObj.defaultDirectory);
    }
    getAggregateFullNameFromFilePath(filePath) {
        return bundlePathHelper_1.BundlePathHelper.scanFilePathForAggregateFullName(filePath, this.typeDefObj.defaultDirectory);
    }
    getAggregateMetadataFilePathFromWorkspacePath(filePath) {
        // wave templates have no metadata file, so return the path to the bundle directory
        const bundleName = this.getAggregateFullNameFromFilePath(filePath);
        const pathToWaveTemplates = PathUtil.getPathToDir(filePath, this.typeDefObj.defaultDirectory);
        return path.join(pathToWaveTemplates, bundleName);
    }
    getDefaultAggregateMetadataPath(fullName, defaultSourceDir, bundleFileProperties) {
        // If there is a bundle properties, and there is always only 0 or 1 values here, then check for additional
        // path details to include between the bundleName and fileName.
        const bundle = bundleFileProperties[0];
        if (bundle) {
            const name = bundle.fileName;
            let baseFileName = path.basename(name);
            const additionalPath = bundlePathHelper_1.BundlePathHelper.getExtendedBundlePath(name, fullName);
            if (additionalPath) {
                return path.join(defaultSourceDir, this.typeDefObj.defaultDirectory, fullName, additionalPath, baseFileName);
            }
            else {
                return path.join(defaultSourceDir, this.typeDefObj.defaultDirectory, fullName, baseFileName);
            }
        }
        else {
            return path.join(defaultSourceDir, this.typeDefObj.defaultDirectory, fullName);
        }
    }
    getAggregateFullNameFromSourceMemberName(sourceMemberName) {
        return sourceMemberName.split(path.sep)[0];
    }
    getAggregateFullNameFromWorkspaceFullName(workspaceFullName) {
        return workspaceFullName.split(path.sep)[0];
    }
    getAggregateFullNameFromFileProperty(fileProperty, namespace) {
        return fileProperty.fullName;
    }
    getPathToMdapiSourceDir(aggregateFullName, mdDir) {
        return path.join(mdDir, this.typeDefObj.defaultDirectory, aggregateFullName);
    }
    getRetrievedMetadataPath(fileProperty, retrieveRoot, bundleFileProperties) {
        // This gets called during convert and passes the bundleFileProperties that is needed to resolve the path
        const bundle = bundleFileProperties[0];
        if (bundle) {
            const bundlePath = path.dirname(fileProperty.fileName);
            const fileName = path.basename(bundle.fileName);
            const retrievedMetadataPath = path.join(retrieveRoot, bundlePath, fileName);
            if (srcDevUtil.pathExistsSync(retrievedMetadataPath)) {
                return retrievedMetadataPath;
            }
        }
        return null;
    }
    getOriginContentPathsForSourceConvert(metadataFilePath, workspaceVersion, unsupportedMimeTypes, forceIgnore) {
        // WaveTemplateBundles have no metadata files, so the metadataFilePath is the path to the bundle
        return Promise.resolve(bundlePathHelper_1.BundlePathHelper.getAllNestedBundleContentPaths(metadataFilePath, forceIgnore));
    }
    shouldGetMetadataTranslation() {
        return false;
    }
    getWorkspaceContentFilePath(metadataFilePath, retrievedContentFilePath) {
        const fileName = path.basename(retrievedContentFilePath);
        const bundleName = bundlePathHelper_1.BundlePathHelper.scanFilePathForAggregateFullName(metadataFilePath, this.typeDefObj.defaultDirectory);
        const metadataDir = path.dirname(metadataFilePath);
        // If there is an additional path in the retrieved content then check if it's already included in the metadata path, and if
        // not add it.
        const additionalPath = bundlePathHelper_1.BundlePathHelper.getExtendedBundlePath(retrievedContentFilePath, bundleName);
        if (additionalPath) {
            if (path.basename(metadataDir) !== path.basename(additionalPath)) {
                return path.join(metadataDir, additionalPath, fileName);
            }
        }
        return path.join(metadataDir, fileName);
    }
    sourceMemberFullNameConflictsWithWorkspaceFullName(sourceMemberFullName, workspaceFullName) {
        const aggregateSourceMemberName = this.getAggregateFullNameFromSourceMemberName(sourceMemberFullName);
        const aggregateFullName = this.getAggregateFullNameFromWorkspaceFullName(workspaceFullName);
        return aggregateSourceMemberName === aggregateFullName;
    }
    trackRemoteChangeForSourceMemberName(sourceMemberName) {
        // Wave Templates only report changes in sourcemember at the bundle level
        return true;
    }
    onlyDisplayOneConflictPerAggregate() {
        // Wave Templates behave similar to aura bundles, following same logic
        return true;
    }
    getDisplayPathForLocalConflict(workspaceFilePath) {
        return path.dirname(workspaceFilePath);
    }
    getAggregateFullNameFromComponentFailure(componentFailure) {
        // Wave Templates behave similar to aura bundles, following same logic
        return this.getAggregateFullNameFromFilePath(componentFailure.fileName);
    }
    getAggregateFullNameFromMdapiPackagePath(mdapiPackagePath) {
        const pathElements = mdapiPackagePath.split(path.sep);
        return pathElements[1];
    }
    validateDeletedContentPath(deletedContentPath, contentPaths, metadataRegistry) {
        const wtTypeDef = metadataRegistry.getWaveDefByFileName(deletedContentPath);
        if (wtTypeDef && wtTypeDef.hasMetadata) {
            const otherWaveTemplateFilesExist = contentPaths.some(contentPath => !contentPath.endsWith(wtTypeDef.fileSuffix));
            if (otherWaveTemplateFilesExist) {
                const err = new Error();
                err.message = messages.getMessage('MissingContentFile', deletedContentPath);
                err.name = 'Missing Content File';
                throw err;
            }
        }
    }
    getWorkspaceFullNameFromComponentFailure(componentFailure) {
        return this.getFullNameFromFilePath(componentFailure.fileName);
    }
    parseSourceMemberForMetadataRetrieve(sourceMemberName, sourceMemberType, isNameObsolete) {
        const fullName = this.getAggregateFullNameFromSourceMemberName(sourceMemberName);
        return { fullName, type: sourceMemberType, isNameObsolete };
    }
    getMdapiFormattedContentFileName(originContentPath, aggregateFullName) {
        // WaveTemplateBundles bundles can have nested sub-directories.
        // In order maintain the bundle structure in the mdapi formatted directory,
        // return the path from the bundle directory to the file
        const pathToBundle = PathUtil.getPathToDir(originContentPath, aggregateFullName);
        return path.relative(pathToBundle, originContentPath);
    }
    handleSlashesForSourceMemberName(sourceMemberFullName) {
        return PathUtil.replaceForwardSlashes(sourceMemberFullName);
    }
    sourceMemberFullNameCorrespondsWithWorkspaceFullName(sourceMemberFullName, workspaceFullName) {
        return sourceMemberFullName === workspaceFullName;
    }
}
exports.WaveTemplateBundleMetadataType = WaveTemplateBundleMetadataType;

//# sourceMappingURL=waveTemplateBundleMetadataType.js.map
