/*
 * Copyright (c) 2016, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
/**
 * Contains config details and operations meant for package subscriber orgs.
 * This could potentially include test orgs used by CI process for testing packages,
 * and target subscriber orgs.
 **/
const path = require('path');
const Promise = require('bluebird');
const _ = require('lodash');
const util = require('util');
const logger = require(path.join('..', 'logApi'));
const messages = require(path.join(__dirname, '..', 'messages'))();
const pkgUtils = require(path.join(__dirname, '..', 'package', 'packageUtils'));
const DEFAULT_POLL_INTERVAL_MILLIS = 5000;
const REPLICATION_POLLING_INTERVAL_MILLIS = 10000;
const DEFAULT_MAX_RETRIES = 0;
const RETRY_MINUTES_IN_MILLIS = 60000;
const DEFAULT_PUBLISH_WAIT_MIN = 0;
/**
 * Private utility to parse out errors from PackageInstallRequest as a user-readable string.
 */
const readInstallErrorsAsString = function (request) {
    if (request.Errors && request.Errors.errors) {
        const errorsArray = request.Errors.errors;
        const len = errorsArray.length;
        if (len > 0) {
            let errorMessage = 'Installation errors: ';
            for (let i = 0; i < len; i++) {
                errorMessage += `\n${(i + 1)}) ${errorsArray[i].message}`;
            }
            return errorMessage;
        }
    }
    return '<empty>';
};
class PackageInstallCommand {
    constructor(stdinPrompt) {
        this.pollIntervalMillis = DEFAULT_POLL_INTERVAL_MILLIS;
        this.replicationPollIntervalMillis = REPLICATION_POLLING_INTERVAL_MILLIS;
        this.maxRetries = DEFAULT_MAX_RETRIES;
        this.allPackageVersionId = null;
        this.installationKey = null;
        this.publishwait = DEFAULT_PUBLISH_WAIT_MIN;
        this.logger = logger.child('PackageInstallCommand');
        this.stdinPrompt = stdinPrompt;
    }
    poll(context, id, retries) {
        this.org = context.org;
        this.configApi = this.org.config;
        this.force = this.org.force;
        return this.force.toolingRetrieve(this.org, 'PackageInstallRequest', id).then((request) => {
            switch (request.Status) {
                case 'SUCCESS':
                    return request;
                case 'ERROR': {
                    const err = readInstallErrorsAsString(request);
                    this.logger.error('Encountered errors installing the package!', err);
                    throw new Error(err);
                }
                default:
                    if (retries > 0) {
                        // Request still in progress.  Just log a message and move on. Server will be polled again.
                        this.logger.log(messages.getMessage('installStatus', request.Status, 'packaging'));
                        return Promise.delay(this.pollIntervalMillis).then(() => (this.poll(context, id, retries - 1)));
                    }
                    else {
                        // timed out
                    }
            }
            return request;
        });
    }
    _waitForApvReplication(remainingRetries) {
        return __awaiter(this, void 0, void 0, function* () {
            const QUERY_NO_KEY = 'SELECT Id, SubscriberPackageId FROM SubscriberPackageVersion' +
                ` WHERE Id ='${this.allPackageVersionId}'`;
            const QUERY_W_KEY = 'SELECT Id, SubscriberPackageId FROM SubscriberPackageVersion' +
                ` WHERE Id ='${this.allPackageVersionId}' AND InstallationKey ='${this.installationKey}'`;
            let queryResult = null;
            try {
                queryResult = yield this.force.toolingQuery(this.org, util.format(QUERY_W_KEY));
            }
            catch (e) {
                // Check first for Implementation Restriction error that is enforced in 214, before it was possible to query
                // against InstallationKey, otherwise surface the error.
                if (pkgUtils.isErrorFromSPVQueryRestriction(e)) {
                    queryResult = yield this.force.toolingQuery(this.org, util.format(QUERY_NO_KEY));
                }
                else {
                    if (!pkgUtils.isErrorPackageNotAvailable(e)) {
                        throw e;
                    }
                }
            }
            if (queryResult && queryResult.records && queryResult.records.length > 0) {
                return queryResult.records[0];
            }
            if (remainingRetries <= 0) {
                throw new Error(messages.getMessage('errorApvIdNotPublished', [], 'package_install'));
            }
            return Promise.delay(this.replicationPollIntervalMillis).then(() => {
                this.logger.log(messages.getMessage('publishWaitProgress', [], 'package_install'));
                return this._waitForApvReplication(remainingRetries - 1);
            });
        });
    }
    /**
     * This installs a package version into a target org.
     * @param context: heroku context
     * @returns {*|promise}
     */
    execute(context) {
        return __awaiter(this, void 0, void 0, function* () {
            this.org = context.org;
            this.configApi = this.org.config;
            this.force = this.org.force;
            // either of the id or package flag is required, not both at the same time
            if ((!context.flags.id && !context.flags.package) ||
                (context.flags.id && context.flags.package)) {
                const idFlag = context.command.flags.find(x => x.name === 'id');
                const packageFlag = context.command.flags.find(x => x.name === 'package');
                throw new Error(messages.getMessage('errorRequiredFlags', [
                    `--${idFlag.name} (-${idFlag.char})`,
                    `--${packageFlag.name}`
                ], 'package_install'));
            }
            let apvId;
            if (context.flags.id) {
                apvId = context.flags.id;
            }
            else if (context.flags.package) {
                // look up the alias only when it's not a 04t
                apvId = (context.flags.package.startsWith(pkgUtils.BY_LABEL.SUBSCRIBER_PACKAGE_VERSION_ID.prefix)) ?
                    context.flags.package :
                    pkgUtils.getPackageIdFromAlias(context.flags.package, this.force);
            }
            // validate whatever is set as the apvId, even if that might be a bunk alias
            try {
                pkgUtils.validateId(pkgUtils.BY_LABEL.SUBSCRIBER_PACKAGE_VERSION_ID, apvId);
            }
            catch (err) {
                throw new Error(messages.getMessage('invalidIdOrPackage', apvId, 'package_install'));
            }
            this.allPackageVersionId = apvId;
            this.maxRetries = _.isNil(context.flags.wait) ? this.maxRetries : ((RETRY_MINUTES_IN_MILLIS / this.pollIntervalMillis) * context.flags.wait);
            // Be careful with the fact that cmd line flags are NOT camel cased: flags.installationkey
            this.installationKey = _.isNil(context.flags.installationkey) ? this.installationKey : context.flags.installationkey;
            if (process.argv.indexOf('-p') > -1) {
                logger.warnUser(context, messages.getMessage('deprecatePublishWaitShortName', [], 'package_install'));
            }
            this.publishwait = _.isNil(context.flags.publishwait) ? this.publishwait : context.flags.publishwait;
            const apiVersion = this.configApi.getApiVersion();
            if (apiVersion < 36) {
                throw new Error('This command is supported only on API versions 36.0 and higher');
            }
            const publishWaitRetries = Math.ceil((parseInt(this.publishwait) * 60 * 1000) / parseInt(this.replicationPollIntervalMillis));
            yield this._waitForApvReplication(publishWaitRetries);
            // If the user is installing an unlocked package with external sites (RSS/CSP) then
            // inform and prompt the user of these sites for acknowledgement.
            const externalSiteData = yield this._getExternalSites(this.allPackageVersionId);
            let enableExternalSites = false;
            if (externalSiteData.trustedSites && externalSiteData.trustedSites.length > 0) {
                const accepted = yield this._promptForTrustedExternalSites(context.flags.noprompt, externalSiteData.trustedSites);
                if (accepted) {
                    enableExternalSites = true;
                }
            }
            // Construct PackageInstallRequest sobject used to trigger package version install.
            const packageInstallRequest = {};
            packageInstallRequest.subscriberPackageVersionKey = this.allPackageVersionId;
            packageInstallRequest.password = this.installationKey; // W-3980736 in the future we hope to change "Password" to "InstallationKey" on the server
            // Add default parameters to input object.
            packageInstallRequest.securityType = 'Full';
            packageInstallRequest.nameConflictResolution = 'Block';
            packageInstallRequest.packageInstallSource = 'U';
            packageInstallRequest.enableRss = enableExternalSites;
            const result = yield this.force.toolingCreate(this.org, 'PackageInstallRequest', packageInstallRequest);
            const packageInstallRequestId = result.id;
            if (_.isNil(packageInstallRequestId)) {
                throw new Error(`Failed to create PackageInstallRequest for: ${this.allPackageVersionId}`);
            }
            return this.poll(context, packageInstallRequestId, this.maxRetries);
        });
    }
    _promptForTrustedExternalSites(noninteractive, externalSites) {
        return __awaiter(this, void 0, void 0, function* () {
            const answer = noninteractive ? 'YES'
                : yield this.stdinPrompt(messages.getMessage('prompt', externalSites.join('\n'), 'package_install'));
            return (answer.toUpperCase() === 'YES' || answer.toUpperCase() === 'Y');
        });
    }
    /**
     * Returns all RSS/CSP external third party websites
     * @param allPackageVersionId
     * @returns {object}
     *
     * {
     *   unlocked: boolean,
     *   trustedSites: [string]
     * }
     */
    _getExternalSites(allPackageVersionId) {
        return __awaiter(this, void 0, void 0, function* () {
            const subscriberPackageValues = {};
            const query_no_key = 'SELECT RemoteSiteSettings, CspTrustedSites ' +
                'FROM SubscriberPackageVersion ' +
                `WHERE Id ='${allPackageVersionId}'`;
            const query_w_key = 'SELECT RemoteSiteSettings, CspTrustedSites ' +
                'FROM SubscriberPackageVersion ' +
                `WHERE Id ='${allPackageVersionId}' AND InstallationKey ='${this.installationKey}'`;
            let queryResult = null;
            try {
                queryResult = yield this.force.toolingQuery(this.org, query_w_key);
            }
            catch (e) {
                // Check first for Implementation Restriction error that is enforced in 214, before it was possible to query
                // against InstallationKey, otherwise surface the error.
                if (pkgUtils.isErrorFromSPVQueryRestriction(e)) {
                    queryResult = yield this.force.toolingQuery(this.org, query_no_key);
                }
                else {
                    throw e;
                }
            }
            if (queryResult.records && queryResult.records.length > 0) {
                const record = queryResult.records[0];
                const rssUrls = record.RemoteSiteSettings.settings.map(rss => rss.url);
                const cspUrls = record.CspTrustedSites.settings.map(csp => csp.url);
                subscriberPackageValues.trustedSites = rssUrls.concat(cspUrls);
            }
            return subscriberPackageValues;
        });
    }
    /**
     * returns a human readable message for a cli output
     * @returns {string}
     */
    getHumanSuccessMessage(result) {
        switch (result.Status) {
            case 'SUCCESS':
                return messages.getMessage(result.Status, [result.SubscriberPackageVersionKey], 'package_install_report');
            case 'TERMINATED':
                return '';
            default:
                return messages.getMessage(result.Status, [result.Id, this.org.name], 'package_install_report');
        }
    }
}
module.exports = PackageInstallCommand;

//# sourceMappingURL=packageInstallCommand.js.map
