/*
 * Copyright (c) 2016, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
// Node
const path = require('path');
const util = require('util');
// Local
const messages = require(path.join(__dirname, '..', 'messages'))();
const CommandRegistry = require(path.join(__dirname, '..', 'commandRegistry'));
const pkgUtils = require(path.join(__dirname, 'packageUtils'));
class PackageVersionUpdateCommand {
    constructor() {
        this.error = null;
    }
    /**
     * Convert the list of command line options to a JSON object that can be used to update an Package2Version entity.
     * @param context
     * @returns {{Id: (string|string|string), Name: (string|string), Description: (string|string), Branch: (string|string), Tag: (string|string), IsReleased: (boolean}}
     * @private
     */
    _updatePackageVersionRequestFromContext(context) {
        const values = {};
        const mapping = {
            package: 'Id',
            installationkey: 'InstallKey',
            versionname: 'Name',
            versiondescription: 'Description'
        };
        const globals = [...new Set(CommandRegistry.GLOBAL_PARAMETERS.map(global => global.name)),
            CommandRegistry.TARGET_DEV_HUB_USERNAME_PARAMETER.name
        ];
        Object.keys(context.flags).forEach((flag) => {
            if (!globals.includes(flag) && !util.isNullOrUndefined(context.flags[flag])) {
                const apiName = mapping[flag] ? mapping[flag] : flag.charAt(0).toUpperCase() + flag.slice(1);
                values[apiName] = context.flags[flag];
            }
        });
        return values;
    }
    execute(context) {
        return this._execute(context)
            .catch((err) => {
            // TODO
            // until package2 is GA, wrap perm-based errors w/ 'contact sfdc' action (REMOVE once package2 is GA'd)
            err = pkgUtils.massageErrorMessage(err);
            throw pkgUtils.applyErrorAction(err);
        });
    }
    _execute(context) {
        return __awaiter(this, void 0, void 0, function* () {
            this.org = context.org;
            this.force = context.org.force;
            const packageVersionId = pkgUtils.getPackageIdFromAlias(context.flags.package, this.force);
            // ID can be an 04t or 05i
            pkgUtils.validateId([pkgUtils.BY_LABEL.SUBSCRIBER_PACKAGE_VERSION_ID, pkgUtils.BY_LABEL.PACKAGE_VERSION_ID], packageVersionId);
            // lookup the 05i ID, if needed
            context.flags.package = yield pkgUtils.getPackageVersionId(packageVersionId, this.force, this.org);
            const request = this._updatePackageVersionRequestFromContext(context);
            return this.force.toolingUpdate(this.org, 'Package2Version', request)
                .then((updateResult) => __awaiter(this, void 0, void 0, function* () {
                if (!updateResult.success) {
                    throw new Error(updateResult.errors);
                }
                // Use the 04t ID for the success messgae
                updateResult.id = yield pkgUtils.getSubscriberPackageVersionId(packageVersionId, this.force, this.org);
                return updateResult;
            }));
        });
    }
    getHumanSuccessMessage(result) {
        return messages.getMessage('humanSuccess', [result.id], 'package_version_update');
    }
}
module.exports = PackageVersionUpdateCommand;

//# sourceMappingURL=packageVersionUpdateCommand.js.map
