"use strict";
/*
 * Copyright (c) 2016, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const varargsCommand_1 = require("./varargsCommand");
// Node
const fs = require('fs');
const path = require('path');
// Thirdparty
const Bluebird = require('bluebird');
const _ = require('lodash');
// Local
const Org = require(path.join(__dirname, 'scratchOrgApi'));
const ScratchOrgInfoApi = require(path.join(__dirname, 'scratchOrgInfoApi'));
const Alias = require(path.join(__dirname, 'alias'));
const almError = require(path.join('..', 'lib', 'almError'));
const envTypes = require(path.join('..', 'lib', 'envTypes'));
const fs_readFile = Bluebird.promisify(fs.readFile);
const messages = require(path.join(__dirname, 'messages'))();
const srcDevUtil = require(path.join('..', 'lib', 'srcDevUtil'));
const scratchOrgInfoGenerator = require(path.join(__dirname, 'scratchOrgInfoGenerator'));
const defaultConnectedAppInfo = require(path.join(__dirname, 'defaultConnectedApp'));
const core_1 = require("@salesforce/core");
const ts_types_1 = require("@salesforce/ts-types");
const _ENV_TYPES = envTypes;
const _DEFAULT_ENV_TYPE = '';
const TOPIC = '/event/OrgLifecycleNotification';
const _getClientSecretError = function () {
    const error = new Error(messages.getMessage('ClientSecretRequired'));
    error.name = 'ClientSecretRequired';
    return error;
};
const _getClientSecret = function (stdinValues, org) {
    if (!_.isNil(stdinValues)) {
        // If the user provided anything via stdin, it has to be a non-null, non-whitespace secret.
        const clientSecret = stdinValues.get('secret');
        if (_.isNil(clientSecret) || _.isNil(clientSecret.trim())) {
            return Promise.reject(_getClientSecretError());
        }
        return Promise.resolve(clientSecret.trim());
    }
    else {
        // The user didn't provide a secret, so get it from the Hub config, if it exists.
        return org.getConfig()
            .then((hubOrgConfig) => {
            let hubOrgClientSecret = hubOrgConfig.clientSecret;
            if (!_.isNil(hubOrgClientSecret)) {
                hubOrgClientSecret = hubOrgClientSecret.trim();
                if (_.isNil(hubOrgClientSecret)) {
                    return Promise.reject(_getClientSecretError());
                }
                return hubOrgClientSecret;
            }
            else if (_.isNil(hubOrgConfig.privateKey)) {
                // The clientSecret and privateKey are both missing, using an access token so use the default connected app
                if (org.usingAccessToken) {
                    return defaultConnectedAppInfo.clientSecret;
                }
                return Promise.reject(_getClientSecretError());
            }
            else {
                // JWT OAuth flow. There'd better be a privateKey in the Hub, or else
                // we're going to have no way to authenticate into the resulting org.
                return null;
            }
        }).catch(err => Promise.reject(err));
    }
};
// A validator function to ensure any options parameters entered by the user adhere
// to a whitelist of valid option settings. Because org:create allows options to be
// input either key=value pairs or within the definition file, this validator is
// executed within the ctor and also after parsing/normalization of the definition file.
const optionsValidator = (key, value, scratchOrgInfoPayload) => {
    if (key.toLowerCase() === 'durationdays') {
        throw almError('unrecognizedScratchOrgOption', 'durationDays');
    }
    if (key.toLowerCase() === 'snapshot') {
        const foundInvalidFields = [];
        OrgCreateCommand.SNAPSHOT_UNSUPPORTED_OPTIONS.forEach((invalidField) => {
            if (scratchOrgInfoPayload.hasOwnProperty(invalidField)) {
                foundInvalidFields.push(invalidField);
            }
        });
        if (foundInvalidFields.length > 0) {
            const msg = messages.getMessage('unsupportedSnapshotOrgCreateOptions', [foundInvalidFields.join(', ')], 'orgSnapshot');
            throw new Error(msg);
        }
    }
};
/**
 * constructs a create command helper
 * @param force - the force api
 * @constructor
 */
class OrgCreateCommand extends varargsCommand_1.default {
    constructor() {
        super('org:create', [optionsValidator]);
        this.workspaceType = _DEFAULT_ENV_TYPE;
    }
    /**
     * secondary validation from the cli interface. this is a protocol style function intended to be represented by other
     * commands
     * @param context - this cli context
     * @returns {Promise}
     */
    validate(context) {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            // validate varargs
            yield _super("validate").call(this, context);
            this.org = context.org;
            this.force = this.org.force;
            const fixedContext = srcDevUtil.fixCliContext(context);
            // Ensure we have an org config input source.
            if (!fixedContext.definitionjson && !fixedContext.definitionfile && !_.get(context, 'args.length')) {
                const message = messages.getMessage('cliForceCreateNoConfig');
                return Promise.reject(new Error(message));
            }
            if (!_.isNil(fixedContext.env) && !Object.hasOwnProperty.call(_ENV_TYPES, fixedContext.env)) {
                const envError = new Error(messages.getMessage('unsupportedValueForEnv', fixedContext.env));
                envError.name = 'UnsupportedValueForCreateEnv';
                return Promise.reject(envError);
            }
            if (!_.isNil(fixedContext.durationdays)) {
                const durationNum = parseFloat(fixedContext.durationdays);
                if (!_.isInteger(durationNum) || !_.inRange(durationNum, 1, 31)) {
                    throw almError('unsupportedValueForDuration', fixedContext.durationdays);
                }
            }
            return this.org.getConfig()
                .then(config => {
                // If the env flag is not specified and the url is not an internal salesforce url.
                if (_.isNil(fixedContext.env) && !srcDevUtil.isInternalUrl(config.instanceUrl)) {
                    fixedContext.env = _ENV_TYPES.sandbox;
                }
                this.workspaceType = fixedContext.env || this.workspaceType;
                return fixedContext;
            })
                .catch(e => {
                if (e.name === 'InvalidProjectWorkspace' && !fixedContext.targetdevhubusername) {
                    const newErrorKey = 'NoWorkspaceOrUser';
                    e.name = newErrorKey;
                    e.message = messages.getMessage('NoWorkspaceOrUser');
                }
                throw e;
            });
        });
    }
    streamProcessor(message) {
        const lifecycleRequestId = _.get(message, 'payload.LifecycleRequestId');
        const status = _.get(message, 'payload.Status');
        if (lifecycleRequestId && (lifecycleRequestId === this.scratchOrgInfoId)) {
            if (status === 'Active') {
                return {
                    payload: lifecycleRequestId,
                    completed: true
                };
            }
            else if (status === 'Error') {
                const errorMessage = messages.getMessage('signupFailed', ts_types_1.getAsString(message, 'message.payload.ErrorCode') ||
                    ts_types_1.getAsString(message, 'message.payload.StatusCode') ||
                    'unknown');
                const error = new Error(errorMessage);
                error.name = 'ScratchOrgInfoError';
                throw error;
            }
        }
        return { completed: false };
    }
    /**
     * This performs three functions.
     * 1) Setup a stream listener for a OrgLifecycleNotification
     * 2) Prepares the cli input for the scratchOrgInfo request
     * 3) Calls the server to create a scratchOrgInfo
     * @returns {Promise}
     * @private
     */
    signupOrgWithStreaming(scratchOrgInfoApi, cliContext) {
        return __awaiter(this, void 0, void 0, function* () {
            const scratchOrgInfo = yield this._getScratchOrgInfo(this.org, cliContext);
            const coreOrg = yield core_1.Org.create(this.org.name);
            // Set the API version on the org connection so the streaming client uses the correct version.
            coreOrg.getConnection().setApiVersion(this.force.config.getApiVersion());
            const options = new core_1.DefaultStreamingOptions(coreOrg, TOPIC, this.streamProcessor.bind(this));
            if (cliContext.wait) {
                options.setSubscribeTimeout(new core_1.Time(cliContext.wait, core_1.TIME_UNIT.MINUTES));
            }
            if (scratchOrgInfo.snapshot && !cliContext.wait) {
                // snapshot creates may take longer than regular as snapshots are not pooled
                // TODO: (a) figure out pooling for snapshots and (b) move org:create to async and
                // develop post-create command to finalize org setup (org preferences, eg)
                const defaultTimeout = options.subscribeTimeout;
                options.setSubscribeTimeout(new core_1.Time(defaultTimeout.milliseconds * 2, core_1.TIME_UNIT.MILLISECONDS));
            }
            try {
                const asyncStatusClient = yield core_1.StreamingClient.init(options);
                yield asyncStatusClient.handshake();
                this.logger.debug('Create org handshake complete');
                return yield asyncStatusClient.subscribe(() => __awaiter(this, void 0, void 0, function* () {
                    const requestResponse = yield scratchOrgInfoApi.request(scratchOrgInfo);
                    this.scratchOrgInfoId = requestResponse.id;
                }));
            }
            catch (e) {
                if (this.scratchOrgInfoId) {
                    // If any error occurs and we happen to already have a scratchOrgId lets query one more time
                    // for the org status. Just in case the streaming world wasn't helpful.
                    try {
                        const orgInfo = yield scratchOrgInfoApi.retrieveScratchOrgInfo(this.scratchOrgInfoId);
                        if (orgInfo) {
                            if (orgInfo.Status === 'Active') {
                                return this.scratchOrgInfoId;
                            }
                        }
                    }
                    catch (retrieveError) {
                        // If we get any errors from the secondary query throw the outer error and log the inner error.
                        if (retrieveError.name === 'RemoteOrgSignupFailed') {
                            throw retrieveError;
                        }
                        else {
                            // This will likely never happen. Unknown signup status....
                            this.logger.debug(retrieveError);
                            throw e;
                        }
                    }
                }
                throw e;
            }
        });
    }
    signupOrgWithPolling(scratchOrgInfoApi, cliContext) {
        return __awaiter(this, void 0, void 0, function* () {
            const client = yield core_1.PollingClient.init(new core_1.DefaultPollingOptions(() => __awaiter(this, void 0, void 0, function* () {
                if (this.scratchOrgInfoId) {
                    this.logger.debug(`polling client this.org.name: ${this.org.name}`);
                    const result = yield scratchOrgInfoApi.retrieveScratchOrgInfo(this.scratchOrgInfoId);
                    this.logger.debug(`polling client result: ${result}`);
                    if (result.Status === 'Active') {
                        return {
                            completed: true,
                            payload: {
                                Status: result.Status,
                                Id: result.I
                            }
                        };
                    }
                }
                return { completed: false };
            })));
            const scratchOrgInfo = yield this._getScratchOrgInfo(this.org, cliContext);
            const scratchOrgInfoRequestResult = yield scratchOrgInfoApi.request(scratchOrgInfo);
            this.scratchOrgInfoId = scratchOrgInfoRequestResult.id;
            const result = yield client.subscribe();
            return result.Id;
        });
    }
    /**
     * executes the command. this is a protocol style function intended to be represented by other commands.
     * @param cliContext - the cli context
     * @param stdinValues - param values obtained from stdin
     * @returns {Promise}
     */
    execute(cliContext, stdinValues) {
        return __awaiter(this, void 0, void 0, function* () {
            const scratchOrgInfoApi = new ScratchOrgInfoApi(this.force, this.org);
            const scratchOrgApi = new Org(this.force);
            const configAggregator = yield core_1.ConfigAggregator.create();
            const fallbackPolling = configAggregator.getPropertyValue(core_1.Config.USE_BACKUP_POLLING_ORG_CREATE);
            this.logger.debug(`fallbackPolling: ${!!fallbackPolling}`);
            let scratchOrgInfoId;
            try {
                scratchOrgInfoId = yield this.signupOrgWithStreaming(scratchOrgInfoApi, cliContext);
            }
            catch (e) {
                if (e.name) {
                    /**
                     * If the streaming client times out doing a handshake then we will default to polling
                     * only if the the config option is enabled.
                     */
                    if (fallbackPolling && e.name === core_1.StreamingTimeoutErrorType.HANDSHAKE) {
                        scratchOrgInfoId = yield this.signupOrgWithPolling(scratchOrgInfoApi, cliContext);
                    }
                    else if (e.name === core_1.StreamingTimeoutErrorType.SUBSCRIBE) {
                        // This happens if the stream listener is aborted.
                        // We will report back to the user, understanding there is no auth stored for this org because of the
                        // timeout. Without a password or token this org is basically in an abandoned state.
                        if (this.scratchOrgInfoId) {
                            throw almError('genericTimeoutMessage', [], 'genericTimeoutCommandWaitMessageAction', [`sfdx force:data:soql:query -q "SELECT Status FROM ScratchOrgInfo WHERE Id='${this.scratchOrgInfoId}'"`]);
                        }
                        else {
                            throw almError('genericTimeoutMessage', [], 'genericTimeoutWaitMessageAction');
                        }
                    }
                    else {
                        throw e;
                    }
                }
            }
            const scratchOrgInfoResult = yield scratchOrgInfoApi.retrieveScratchOrgInfo(scratchOrgInfoId);
            const clientSecret = yield _getClientSecret(stdinValues, this.org);
            scratchOrgApi.alias = cliContext.setalias;
            const orgData = yield scratchOrgInfoApi.processScratchOrgInfoResult(scratchOrgInfoResult, clientSecret, scratchOrgApi, !!cliContext.setdefaultusername);
            if (cliContext.setalias) {
                yield Alias.set(cliContext.setalias, scratchOrgApi.getName());
            }
            return { orgId: orgData.orgId, username: scratchOrgApi.getName() };
        });
    }
    // Returns a valid signup json object
    _getScratchOrgInfo(masterOrg, context) {
        // Varargs input overrides definitionjson (-j option; hidden/deprecated)
        const definitionJson = context.definitionjson ? JSON.parse(context.definitionjson) : {};
        const orgConfigInput = Object.assign(definitionJson, this.keyValuePairs || {});
        let scratchOrgInfoPayloadPromise;
        // the -f option
        if (context.definitionfile) {
            scratchOrgInfoPayloadPromise = fs_readFile(context.definitionfile)
                // definitionjson and varargs override file input
                .then(defFileContents => Object.assign({}, JSON.parse(defFileContents), orgConfigInput))
                .catch(err => {
                err = srcDevUtil.processReadAndParseJsonFileError(err, context.definitionfile);
                return Promise.reject(err);
            });
        }
        else {
            scratchOrgInfoPayloadPromise = Bluebird.resolve(orgConfigInput);
        }
        return scratchOrgInfoPayloadPromise
            .then((scratchOrgInfoPayload) => {
            // scratchOrgInfoPayload must be heads down camelcase.
            const upperCaseKey = srcDevUtil.findUpperCaseKeys(scratchOrgInfoPayload);
            if (upperCaseKey) {
                throw almError('InvalidJsonCasing', [upperCaseKey, JSON.stringify(scratchOrgInfoPayload, null, 4)]);
            }
            // Now run the fully resolved user input against the validator
            _.forEach(scratchOrgInfoPayload, (value, key) => optionsValidator(key, value, scratchOrgInfoPayload));
            // the -i option
            if (!_.isNil(context.clientid)) {
                scratchOrgInfoPayload.connectedAppConsumerKey = context.clientid;
            }
            // the -d option
            if (!_.isNil(context.durationdays)) {
                scratchOrgInfoPayload.durationDays = context.durationdays;
            }
            // Ignore ancestor ids only when 'nonamespace' or 'noancestors' options are specified
            let ignoreAncestorIds = (context.nonamespace || context.noancestors || false);
            return scratchOrgInfoGenerator.generateScratchOrgInfo(masterOrg, scratchOrgInfoPayload, 'scratch', context.nonamespace, ignoreAncestorIds);
        });
    }
    parseSignupErrorCode(err) {
        const messageToParse = err.name + err.errorCode + err.message;
        return messageToParse.match(/[A-Z]-[0-9]{4}/);
    }
    /**
     * returns a human readable message for cli output
     * @param org - the result oif execute
     * @returns {string}
     */
    getHumanSuccessMessage(org) {
        return messages.getMessage('createOrgCommandSuccess', [org.orgId, org.username]);
    }
    /**
     * returns a human readable error message for cli output
     * @returns {string}
     */
    getHumanErrorMessage(err) {
        let message;
        try {
            // try and get a message associated with an embedded error code.
            const errorCode = this.parseSignupErrorCode(err);
            if (!_.isNil(errorCode) && errorCode.length > 0) {
                message = messages.getMessage(errorCode[0], [this.force.config.getWorkspaceConfigFilename()], 'signup');
            }
            // didn't have a salesforce formatted error code. try a lookup from just the error code
            else {
                message = messages.getMessage(err.errorCode, [this.force.config.getWorkspaceConfigFilename()], 'signup');
            }
        }
        // getMessage throws an error when a key is not found.
        catch (e) {
            // Use the standard error message
            message = null;
        }
        return message;
    }
}
OrgCreateCommand.SNAPSHOT_UNSUPPORTED_OPTIONS = ['features', 'orgPreferences', 'edition', 'sourceOrg', 'settingsPath',
    'releaseVersion', 'language'];
module.exports = OrgCreateCommand;

//# sourceMappingURL=scratchOrgCreateCommand.js.map
