/*
 * Copyright (c) 2016, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
// Node
const fs = require('fs-extra');
const os = require('os');
const util = require('util');
const path = require("path");
// Thirdparty
const Promise = require('bluebird');
const _ = require('lodash');
const mkdirp = require('mkdirp');
const js2xmlparser = require('js2xmlparser');
// Local
const MdapiDeployApi = require(path.join(__dirname, 'mdapiDeployApi'));
Promise.promisifyAll(fs);
/** This is the contents of the package.xml that we will use when we deploy settings to a scratch org. */
const _settingsPackageFileContents = '<?xml version="1.0" encoding="UTF-8"?>' +
    '<Package xmlns="http://soap.sforce.com/2006/04/metadata">' +
    '    <types>' +
    '%s' +
    '        <name>Settings</name>' +
    '    </types>' +
    '    <version>%s</version>' +
    '</Package>';
const _settingsPackageMemberTemplate = '        <members>*</members>';
/**
 * Helper class for dealing with the settings that are defined in a scratch definition file.  This class knows how to extract the
 * settings from the definition, how to expand them into a MD directory and how to generate a package.xml.
 */
class SettingsGenerator {
    /** extract the settings from the scratch def file, if they are present. */
    extract(scratchDef) {
        this.settingData = _.get(scratchDef, 'settings');
        if (this.hasSettings()) {
            //TODO, this is where we will validate the settings.
            //See W-5068155
        }
    }
    /** True if we are currently tracking setting data. */
    hasSettings() {
        return !_.isEmpty(this.settingData);
    }
    /** Create temporary deploy directory used to upload the scratch org shape.
     * This will create the dir, generate package.xml and all of the .setting files.
     * @returns {*|Promise.<{}>}
     */
    createDeployDir(apiVersion) {
        // The root of our package
        const destRoot = path.join(os.tmpdir(), 'shape');
        const settingsDir = path.join(destRoot, 'settings');
        const packageFile = path.join(destRoot, 'package.xml');
        return fs.exists(destRoot)
            // Create a deploy directory
            .then((present) => {
            if (present) {
                fs.removeSync(destRoot);
            }
            return mkdirp.sync(settingsDir);
        })
            // Generate all .settings files into our temporary deploy directory for eash JSON setting
            .then(() => {
            const files = [];
            Object.keys(this.settingData).forEach(item => {
                const value = _.get(this.settingData, item);
                const typeName = this.cap(item);
                const fname = typeName.replace("Settings", "");
                const f = this._createSettingsFileContent(typeName, value);
                files.push(fs.writeFile(path.join(settingsDir, fname + '.settings'), f));
            });
            return Promise.all(files);
        })
            // Create our package.xml file in the settings root
            .then(() => {
            let memberReferences = '';
            Object.keys(this.settingData).forEach(item => {
                const typeName = this.cap(item).replace("Settings", "");
                memberReferences += '<members>' + typeName + "</members>";
            });
            return fs.writeFile(packageFile, util.format(_settingsPackageFileContents, memberReferences, apiVersion));
        })
            .then(() => destRoot);
    }
    _createSettingsFileContent(name, json) {
        if (name == `OrgPreferenceSettings`) {
            //this is a stupid format
            let res = '<?xml version="1.0" encoding="UTF-8"?>\n'
                + '<OrgPreferenceSettings xmlns="http://soap.sforce.com/2006/04/metadata">\n';
            Object.keys(json).forEach(pref => {
                res += '\t<preferences>\n';
                res += '\t\t<settingName>' + this.cap(pref) + '</settingName>\n';
                res += '\t\t<settingValue>' + _.get(json, pref) + '</settingValue>\n';
                res += '\t</preferences>\n';
            });
            res += '</OrgPreferenceSettings>';
            return res;
        }
        else {
            return js2xmlparser.parse(name, json);
        }
    }
    cap(s) {
        return s.charAt(0).toUpperCase() + s.substring(1);
    }
}
module.exports = SettingsGenerator;

//# sourceMappingURL=scratchOrgSettingsGenerator.js.map
