"use strict";
/*
 * Copyright (c) 2018, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const util = require("util");
const path = require("path");
const moment = require("moment");
const logger = require(path.join(__dirname, '..', 'logApi'));
const messages = require(path.join(__dirname, '..', '..', 'lib', 'messages'))();
exports.ORG_SNAPSHOT_FIELDS = ['Id', 'SnapshotName', 'Description', 'Status', 'SourceOrg',
    'CreatedDate', 'LastModifiedDate', 'ExpirationDate', 'LastClonedDate', 'LastClonedById',
    'Error'];
const DATETIME_FORMAT = 'YYYY-MM-DD HH:mm:ss';
const DATE_FORMAT = 'YYYY-MM-DD';
exports.ORG_SNAPSHOT_COLUMNS = [
    { key: 'Id', label: 'Id' },
    { key: 'SnapshotName', label: 'SnapshotName' },
    { key: 'Status', label: 'Status' },
    { key: 'SourceOrg', label: 'Source Org Id' },
    { key: 'CreatedDate', label: 'Created Date',
        format: (value) => value ? moment(value).format(DATETIME_FORMAT) : ''
    },
    { key: 'LastModifiedDate', label: 'Last Modified Date',
        format: (value) => value ? moment(value).format(DATETIME_FORMAT) : ''
    },
    { key: 'ExpirationDate', label: 'Expiration Date',
        format: (value) => value ? moment(value).format(DATE_FORMAT) : ''
    },
    { key: 'LastClonedDate', label: 'Last Cloned Date',
        format: (value) => value ? moment(value).format(DATETIME_FORMAT) : ''
    },
    { key: 'LastClonedById', label: 'Last Cloned By Id' }
];
const LIST_QUERY = `SELECT ${exports.ORG_SNAPSHOT_FIELDS.join(',')} FROM OrgSnapshot ORDER BY CreatedDate`;
const GET_QUERY_BY_ID = `SELECT ${exports.ORG_SNAPSHOT_FIELDS.join(',')} FROM OrgSnapshot WHERE Id=\'%s\'`;
const GET_QUERY_BY_NAME = `SELECT ${exports.ORG_SNAPSHOT_FIELDS.join(',')} FROM OrgSnapshot WHERE SnapshotName LIKE \'%s\'`;
exports.PERM_QUERY = 'SELECT Id FROM OrgSnapshot LIMIT 1';
const CONTENT = 'metadatadata';
const ORG_SNAPSHOT = 'OrgSnapshot';
const ORG_SNAPSHOT_KEY_PREFIX = '0Oo';
const ORG_SNAPSHOT_NOT_SUPPORTED = `'${ORG_SNAPSHOT}' is not supported`;
/**
 * Org Snapshot API Implementation.
 */
class OrgSnapshotApiImpl {
    static create(org) {
        return __awaiter(this, void 0, void 0, function* () {
            const api = new OrgSnapshotApiImpl(org);
            yield api.checkOrgSnapshotPerm();
            return api;
        });
    }
    // use create()
    constructor(org) {
        this.devHubOrg = org;
        this.force = org.force;
        this.logger = logger.child('OrgSnapshotApi');
    }
    /**
     * Create OrgSnapshot record and org export.
     *
     * @param {OrgSnapshotRequest} request
     * @returns {Promise<OrgSnapshot>}
     */
    create(request) {
        return __awaiter(this, void 0, void 0, function* () {
            request.Content = request.Content || CONTENT;
            let createResult;
            try {
                createResult = yield this.force.create(this.devHubOrg, ORG_SNAPSHOT, request);
            }
            catch (err) {
                this.checkForNotSupported(err);
            }
            if (!createResult.success) {
                throw new Error(createResult.errors);
            }
            // retrieve to show status
            return this.get(createResult.id);
        });
    }
    /**
     * Delete OrgSnapshot record and underlying export.
     *
     * @param {string} orgSnapshotIdOrName
     * @returns {Promise<OrgSnapshot>}
     */
    delete(orgSnapshotIdOrName) {
        return __awaiter(this, void 0, void 0, function* () {
            const orgSnapshotId = yield this.getOrgSnapshotId(orgSnapshotIdOrName);
            let deleteResult;
            try {
                deleteResult = yield this.force.delete(this.devHubOrg, ORG_SNAPSHOT, orgSnapshotId);
            }
            catch (err) {
                this.checkForNotSupported(err);
            }
            if (!deleteResult.success) {
                const errors = deleteResult.errors && deleteResult.errors.length > 0 ? deleteResult.errors.join(', ') : 'Unknown error';
                throw Error(`Unable to delete ${ORG_SNAPSHOT} with name '${orgSnapshotIdOrName}': ${errors}`);
            }
            return deleteResult;
        });
    }
    /**
     * Get OrgSnapshot by given ID or name.
     *
     * @param {string} orgSnapshotIdOrName
     * @returns {Promise<OrgSnapshot>}
     */
    get(orgSnapshotIdOrName) {
        return __awaiter(this, void 0, void 0, function* () {
            const query = orgSnapshotIdOrName.startsWith(ORG_SNAPSHOT_KEY_PREFIX) ? GET_QUERY_BY_ID : GET_QUERY_BY_NAME;
            let queryResult;
            try {
                queryResult = yield this.force.query(this.devHubOrg, util.format(query, orgSnapshotIdOrName));
            }
            catch (err) {
                this.checkForNotSupported(err);
            }
            if (!queryResult.records || !queryResult.records[0]) {
                throw Error(`${ORG_SNAPSHOT} with ID or name '${orgSnapshotIdOrName}' not found`);
            }
            return queryResult.records[0];
        });
    }
    /**
     * Get OrgSnapshot records.
     *
     * @returns {Promise<OrgSnapshot[]>}
     */
    list() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const queryResult = yield this.force.query(this.devHubOrg, LIST_QUERY);
                return queryResult.records ? queryResult.records : [];
            }
            catch (err) {
                this.checkForNotSupported(err);
            }
        });
    }
    /**
     * Returns name-value pairs of column to data
     * @param result
     * @returns {ColumnData[]}
     */
    mapDataToLabel(result) {
        return [
            { name: 'Id', value: result.Id },
            { name: 'Snapshot Name', value: result.SnapshotName },
            { name: 'Description', value: result.Description },
            { name: 'Status', value: result.Status },
            { name: 'Source Org', value: result.SourceOrg },
            { name: 'Expiration Date', value: result.ExpirationDate ? moment(result.ExpirationDate).format(DATE_FORMAT) : '' },
            { name: 'Last Cloned Date', value: result.LastClonedDate ? moment(result.LastClonedDate).format(DATETIME_FORMAT) : '' },
            { name: 'Last Cloned By', value: result.LastClonedById },
            { name: 'Created Date', value: moment(result.CreatedDate).format(DATETIME_FORMAT) },
            { name: 'Last Modified Date', value: moment(result.LastModifiedDate).format(DATETIME_FORMAT) }
        ];
    }
    getOrgSnapshotId(orgSnapshotIdOrName) {
        return __awaiter(this, void 0, void 0, function* () {
            if (orgSnapshotIdOrName.startsWith(ORG_SNAPSHOT_KEY_PREFIX)) {
                return orgSnapshotIdOrName;
            }
            else {
                // retrieve ID
                const record = yield this.get(orgSnapshotIdOrName);
                return record.Id;
            }
        });
    }
    checkOrgSnapshotPerm() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.force.query(this.devHubOrg, exports.PERM_QUERY);
            }
            catch (err) {
                this.checkForNotSupported(err);
            }
            return Promise.resolve(null);
        });
    }
    // inspects if error is lack of OrgSnapshot perm
    checkForNotSupported(err) {
        if (err.message && err.message.includes(ORG_SNAPSHOT_NOT_SUPPORTED)) {
            throw new Error(messages.getMessage('snapshotNotEnabled', [], 'orgSnapshot'));
        }
        else {
            throw err;
        }
    }
}
exports.OrgSnapshotApiImpl = OrgSnapshotApiImpl;

//# sourceMappingURL=orgSnapshotApi.js.map
