/*
 * Copyright (c) 2016, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const Promise = require("bluebird");
const orgSnapshotApi_1 = require("./orgSnapshotApi");
const Org = require("../../lib/scratchOrgApi.js");
const messages = require(path.join(__dirname, '..', '..', 'lib', 'messages'))();
const logger = require(path.join(__dirname, '..', 'logApi'));
/**
 * Create a snapshot of given scratch source org.
 */
class OrgSnapshotCreateCommand {
    constructor() {
        this.logger = logger.child('org:snapshot:create');
    }
    validate(context) {
        // framework handles required params
        return Promise.resolve(context);
    }
    execute(context) {
        // double-check
        if (!context.flags.sourceorg) {
            throw new Error(messages.getMessage('sourceOrgInvalid', [], 'orgSnapshot'));
        }
        if (!context.flags.snapshotname) {
            throw new Error(messages.getMessage('nameInvalid', [], 'orgSnapshot'));
        }
        return this._getSourceOrgId(context)
            .then((orgId) => {
            return orgSnapshotApi_1.OrgSnapshotApiImpl.create(context.org)
                .then(orgSnapshotApi => {
                this.orgSnapshotApi = orgSnapshotApi;
                return this.orgSnapshotApi.create({
                    SourceOrg: orgId,
                    SnapshotName: context.flags.snapshotname,
                    Description: context.flags.description
                });
            });
        });
    }
    // resolve sourceorg to orgId
    _getSourceOrgId(context) {
        if (context.flags.sourceorg.startsWith('00D')) {
            return Promise.resolve(context.flags.sourceorg);
        }
        else {
            return Org.create(context.flags.sourceorg, Org.Defaults.USERNAME)
                .then(org => org.getConfig())
                .then(orgConfig => orgConfig.orgId);
        }
    }
    /**
     * returns a human readable message for a cli output
     * @param result - the data representing the Org Snapshot
     * @returns {string}
     */
    getHumanSuccessMessage(result) {
        const data = this.orgSnapshotApi.mapDataToLabel(result);
        if (result.Error) {
            data.push({ name: 'Error', value: result.Error });
        }
        this.logger.styledHeader(this.logger.color.blue('Org Snapshot'));
        this.logger.table(data, { columns: [
                { key: 'name', label: 'Name' },
                { key: 'value', label: 'Value' }
            ] });
        return '';
    }
}
module.exports = OrgSnapshotCreateCommand;

//# sourceMappingURL=orgSnapshotCreateCommand.js.map
