/*
 * Copyright (c) 2018, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = require('path');
const os = require('os');
const util = require('util');
const sourceDeployApi_1 = require("./sourceDeployApi");
const sourceUtil_1 = require("./sourceUtil");
const core_1 = require("@salesforce/core");
const MdapiPushApi = require(path.join('..', '..', 'lib', 'mdapiPushApi'));
const srcDevUtil = require(path.join('..', '..', 'lib', 'srcDevUtil'));
const logger = require(path.join(__dirname, '..', 'logApi'));
const almError = require(path.join(__dirname, '..', 'almError'));
const syncCommandHelper = require(path.join(__dirname, '..', 'syncCommandHelper'));
const MetadataRegistry = require(path.join(__dirname, '..', 'metadataRegistry'));
core_1.Messages.importMessagesDirectory(__dirname);
class SourceApiCommand {
    constructor(isSourceDelete) {
        this.isSourceDelete = isSourceDelete;
        this.logger = logger.child(`source:${this.deploytype}`);
        this.messagesSourcePush = core_1.Messages.loadMessages('salesforce-alm', 'source_push');
        this.commonMsgs = core_1.Messages.loadMessages('salesforce-alm', 'source');
    }
    getPreExecuteMessage({ orgId, username }) {
        return this.commonMsgs.getMessage(`${this.deploytype}CommandCliPreExecute`, [orgId, username]);
    }
    /**
     * Executes the source deploy or push command
     * @param context - the cli context
     * @returns {Promise}
     */
    execute(context) {
        return __awaiter(this, void 0, void 0, function* () {
            const rows = [];
            const projectPath = this.orgApi.config.getProjectPath();
            return MetadataRegistry.initializeMetadataTypeInfos(this.orgApi)
                .then(() => {
                context.unsupportedMimeTypes = []; // for logging unsupported static resource mime types
                context.delete = this.isSourceDelete; // SourceDeployApi is for change sets and MdapiPushApi is for source push
                const deployApi = this.isDeploy() ? new sourceDeployApi_1.SourceDeployApi(this.orgApi) : new MdapiPushApi(this.orgApi);
                return deployApi.doDeploy(context);
            })
                .catch(e => {
                if (e.name === 'SourceConflict') {
                    const error = almError('sourceConflictDetected');
                    e.sourceConflictElements.forEach(sourceElement => syncCommandHelper.createConflictRows(rows, sourceElement, projectPath));
                    error.columns = syncCommandHelper.getColumnMetaInfo(this.commonMsgs);
                    error.result = rows;
                    this.logger.error(this.messagesSourcePush.getMessage('pushCommandConflictMsg'));
                    throw error;
                }
                else if (e.name === 'DeployFailed') {
                    e.failures.forEach((failure) => syncCommandHelper.createDeployFailureRow(rows, failure, projectPath));
                    const error = almError({ keyName: `source${this.deploytype}Failed`, bundle: `source_${this.deploytype}` });
                    error.name = 'DeployFailed';
                    if (!util.isNullOrUndefined(e.outboundFiles)) {
                        const successes = [];
                        e.outboundFiles.forEach((sourceElement) => syncCommandHelper.createDisplayRows(successes, sourceElement, projectPath));
                        error.partialSuccess = successes.length > 0 ? successes : undefined;
                    }
                    error.columns = this._getErrorColumnData();
                    error.result = rows;
                    if (e.message) {
                        error.message = e.message;
                        this.logger.error(error.message);
                    }
                    throw error;
                }
                else if (e.name === 'PollingTimeout') {
                    const errConfig = new core_1.SfdxErrorConfig('salesforce-alm', 'source', `DeployTimeout`);
                    errConfig.setErrorTokens([this.deploytype, context.wait]);
                    throw core_1.SfdxError.create(errConfig);
                }
                else {
                    throw e;
                }
            })
                .then((result) => {
                if (result.userCanceled) {
                    //user canceled the delete by selecting 'n' when prompted.
                    this.userCanceled = result.userCanceled;
                }
                else {
                    result.outboundFiles.forEach((sourceElement) => syncCommandHelper.createDisplayRows(rows, sourceElement, projectPath));
                }
            })
                .then(() => {
                if (!this.userCanceled) {
                    // User elected to continue the deleted so we can display the proper tabular output.
                    this.logger.styledHeader(this.logger.color.blue(this.commonMsgs.getMessage(this.isSourceDelete ? 'deleteCommandHumanSuccess' : `${this.deploytype}CommandHumanSuccess`)));
                    SourceApiCommand.prototype['getColumnData'] = this._getColumnData;
                }
            })
                .then(() => srcDevUtil.logUnsupportedMimeTypeError(context.unsupportedMimeTypes, this.logger, this.orgApi.force))
                .then(() => {
                if (this.userCanceled) {
                    return {};
                }
                if (context.json) {
                    if (this.isDeploy()) {
                        return this.isSourceDelete ? { deletedSource: rows } : { deployedSource: rows };
                    }
                    return { pushedSource: rows };
                }
                return rows;
            });
        });
    }
    /**
     * Validates the source push or deploy command parameters
     * @param context - the cli context
     * @returns {Promise}
     */
    validate(context) {
        return __awaiter(this, void 0, void 0, function* () {
            this.orgApi = context.org;
            this.verbose = context.flags.verbose;
            this.json = context.flags.json;
            this.deploytype = context.deploytype;
            const fixedContext = srcDevUtil.fixCliContext(context);
            // Validate the wait param if set and convert to an integer.
            sourceUtil_1.parseWaitParam(fixedContext);
            if (this.isDeploy()) {
                // verify that exactly one scoping flag is provided
                sourceUtil_1.validateXorFlags(fixedContext);
                // verify that the org is not a scratch org
                yield sourceUtil_1.validateNonSourceTrackedOrg(this.orgApi.name, 'push');
                // verify that the manifest file exists and is readable.
                if (fixedContext.manifest) {
                    yield sourceUtil_1.validateManifestPath(fixedContext.manifest);
                }
            }
            return Promise.resolve(fixedContext);
        });
    }
    isDeploy() {
        return this.deploytype === SourceApiCommand.SOURCE_DEPLOY;
    }
    /**
     * This indicates to index.js that this command should produce tabular output.
     * @returns {*[]}
     */
    _getColumnData() {
        return syncCommandHelper.getColumnMetaInfo(this.commonMsgs, this.isDeploy());
    }
    getHumanErrorMessage(error) {
        if (!error.partialSuccess) {
            return; // Don't get in the way of single table normal errors
        }
        let needNewLine = false;
        if (error.partialSuccess && error.partialSuccess.length > 0) {
            const headerSuccess = this.commonMsgs.getMessage(`${this.deploytype}CommandHumanSuccess`);
            const columnsSuccess = syncCommandHelper.getColumnMetaInfo(this.commonMsgs);
            this.logger.styledHeader(this.logger.color.blue(headerSuccess));
            this.logger.table(error.partialSuccess, { columns: columnsSuccess });
            needNewLine = true;
        }
        if (error.result.length > 0) {
            if (needNewLine) {
                this.logger.log(os.EOL);
            }
            const headerErrors = this.commonMsgs.getMessage(`${this.deploytype}CommandHumanError`);
            this.logger.styledHeader(this.logger.color.red(headerErrors));
        }
        return;
    }
    _getErrorColumnData() {
        if (this.verbose || this.json) {
            return [
                { key: 'fullName', label: this.commonMsgs.getMessage('fullNameTableColumn') },
                { key: 'type', label: this.commonMsgs.getMessage('typeTableColumn') },
                { key: 'filePath', label: this.commonMsgs.getMessage('workspacePathTableColumn') },
                { key: 'error', label: this.commonMsgs.getMessage('errorColumn') },
                { key: 'lineNumber', label: this.commonMsgs.getMessage('lineNumberColumn') },
                { key: 'columnNumber', label: this.commonMsgs.getMessage('columnNumberColumn') }
            ];
        }
        else {
            return [
                { key: 'filePath', label: this.commonMsgs.getMessage('workspacePathTableColumn') },
                { key: 'error', label: this.commonMsgs.getMessage('errorColumn') }
            ];
        }
    }
}
SourceApiCommand.SOURCE_DEPLOY = 'deploy';
SourceApiCommand.SOURCE_PUSH = 'push';
exports.SourceApiCommand = SourceApiCommand;

//# sourceMappingURL=sourceApiCommand.js.map
