/*
 * Copyright (c) 2018, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const cli_ux_1 = require("cli-ux");
// Node
const path = require('path');
const fsExtra = require('fs-extra');
// Local
const MetadataRegistry = require(path.join(__dirname, '..', 'metadataRegistry'));
const SourceWorkspaceAdapter = require(path.join(__dirname, '..', 'sourceWorkspaceAdapter'));
const syncCommandHelper = require(path.join(__dirname, '..', 'syncCommandHelper'));
const sourceState = require(path.join(__dirname, '..', 'sourceState'));
const core_1 = require("@salesforce/core");
const metadataTypeFactory_1 = require("../metadataTypeFactory");
const sourceDeployApiBase_1 = require("./sourceDeployApiBase");
const SourceUtil = require("./sourceUtil");
const PathUtils = require("../sourcePathUtil");
const os = require("os");
core_1.Messages.importMessagesDirectory(__dirname);
class SourceDeployApi extends sourceDeployApiBase_1.SourceDeployApiBase {
    // @todo we shouldn't cross the command api separation by re-using cli options as dependencies for the api.
    doDeploy(options) {
        return __awaiter(this, void 0, void 0, function* () {
            let aggregateSourceElements = new Map();
            this.isDelete = options.delete;
            this.logger = yield core_1.Logger.child('SourceDeployApi');
            // Only put SWA in stateless mode when sourcepath param is used.
            const mode = options.sourcepath && SourceWorkspaceAdapter.modes.STATELESS;
            this.logger.debug(`mode: ${mode}`);
            const defaultPackagePath = this.orgApi.config.getAppConfig().defaultPackagePath;
            this.swa = new SourceWorkspaceAdapter(this.orgApi, defaultPackagePath, MetadataRegistry, true, mode);
            let tmpOutputDir = yield SourceUtil.createOutputDir('sourceDeploy');
            try {
                if (options.sourcepath) {
                    this.logger.info(`Deploying metadata in sourcepath '${options.sourcepath}' from org: '${this.orgApi.name}'`);
                    aggregateSourceElements = yield SourceUtil.getSourceElementsFromSourcePath(options.sourcepath, this.swa);
                }
                else if (options.manifest) {
                    this.logger.info(`Deploying metadata in manifest '${options.manifest}' from org: '${this.orgApi.name}'`);
                    aggregateSourceElements = yield this._getSourceElementsFromManifest(options.manifest);
                }
                else if (options.metadata) {
                    aggregateSourceElements = yield this._getSourceElementsFromMetadata(options, tmpOutputDir, aggregateSourceElements);
                }
                else {
                    // This should never happen but just a little OC - 'else if' without an 'else'
                    throw core_1.SfdxError.create('salesforce-alm', 'source', 'missingScopeOption');
                }
                const shouldContinue = this.isDelete ? yield this._handleDelete(options.noprompt, aggregateSourceElements, options.sourcepath) : true;
                if (shouldContinue) {
                    if (isNaN(options.wait)) {
                        options.wait = this.force.config.getConfigContent().defaultSrcWaitMinutes;
                    }
                    if (aggregateSourceElements.size > 0) {
                        try {
                            // Create a temp directory
                            options.deploydir = tmpOutputDir;
                            options.ignorewarnings = this.isDelete;
                            yield this._doLocalDelete(aggregateSourceElements);
                            const result = yield this.convertAndDeploy(options, this.swa, aggregateSourceElements, this.isDelete);
                            // process results
                            return yield this._processResults(result, aggregateSourceElements, options.deploydir);
                        }
                        finally {
                            // Remove the sourcePathInfos.json file and delete any temp dirs
                            this.orgApi.getSourcePathInfos().delete();
                            yield SourceUtil.cleanupOutputDir(this.tmpBackupDeletions);
                        }
                    }
                    else {
                        return { outboundFiles: [] };
                    }
                }
                else {
                    return { outboundFiles: [], userCanceled: true };
                }
            }
            finally {
                yield SourceUtil.cleanupOutputDir(tmpOutputDir);
            }
            return { outboundFiles: [] };
        });
    }
    ;
    _getSourceElementsFromMetadata(options, tmpOutputDir, aggregateSourceElements) {
        return __awaiter(this, void 0, void 0, function* () {
            const manifestPath = yield SourceUtil.toManifest(this.orgApi, this.swa.metadataRegistry, options, tmpOutputDir);
            this.logger.info(`Deploying metadata in metadata '${options.metadata}' from org '${this.orgApi.name}'`);
            if (manifestPath) {
                aggregateSourceElements = yield this._getSourceElementsFromManifest(manifestPath);
            }
            else {
                throw core_1.SfdxError.create('salesforce-alm', 'source', 'failedToCreateManifest');
            }
            return aggregateSourceElements;
        });
    }
    _doLocalDelete(ases) {
        return __awaiter(this, void 0, void 0, function* () {
            this.tmpBackupDeletions = yield SourceUtil.createOutputDir('sourceDelete');
            ases.forEach((ase) => {
                ase.getPendingDeletedWorkspaceElements().forEach(we => fsExtra.copySync(we.getSourcePath(), path.join(this.tmpBackupDeletions, path.basename(we.getSourcePath()))));
                ase.commitDeletes([]);
                PathUtils.cleanEmptyDirs(path.dirname(ase.getMetadataFilePath()));
            });
        });
    }
    _handleDelete(noprompt, ases, sourcepath) {
        return __awaiter(this, void 0, void 0, function* () {
            let pendingDelPathsForPrompt = [];
            ases.forEach((ase) => {
                ase.getWorkspaceElements().some((we) => {
                    const type = we.getMetadataName();
                    const sourceMemberMetadataType = metadataTypeFactory_1.MetadataTypeFactory.getMetadataTypeFromMetadataName(type, this.swa.metadataRegistry);
                    const shouldDeleteWorkspaceAggregate = sourceMemberMetadataType.shouldDeleteWorkspaceAggregate(type);
                    if (shouldDeleteWorkspaceAggregate) {
                        ase.markForDelete();
                        return true;
                    }
                    else { // the type is decomposed and we only want to delete components of an aggregate element
                        const sourcepaths = sourcepath.split(',');
                        if (sourcepaths.some(sp => we.getSourcePath().includes(sp.trim()))) {
                            we.setState(sourceState.DELETED);
                            ase.addPendingDeletedWorkspaceElement(we);
                        }
                    }
                });
                pendingDelPathsForPrompt = pendingDelPathsForPrompt.concat(ase.getPendingDeletedWorkspaceElements().map(el => `${os.EOL}${el.getSourcePath()}`));
            });
            if (noprompt || pendingDelPathsForPrompt.length === 0) {
                return true;
            }
            return this._handlePrompt(pendingDelPathsForPrompt);
        });
    }
    ;
    _handlePrompt(pathsToPrompt) {
        return __awaiter(this, void 0, void 0, function* () {
            // @todo this prompt should not be in the API. Need to remove.
            const messages = core_1.Messages.loadMessages('salesforce-alm', 'source_delete');
            const promptMessage = messages.getMessage('sourceDeletePrompt', [pathsToPrompt]);
            const answer = yield cli_ux_1.default.prompt(promptMessage);
            return answer.toUpperCase() === 'YES' || answer.toUpperCase() === 'Y';
        });
    }
    _getSourceElementsFromManifest(optionsManifest) {
        return __awaiter(this, void 0, void 0, function* () {
            const typeNamePairs = yield SourceUtil.parseToManifestEntriesArray(optionsManifest);
            return this._parseTypeNamePairs(typeNamePairs, this.swa.getAggregateSourceElements(false));
        });
    }
    ;
    _parseTypeNamePairs(typeNamePairs, sourceElements) {
        let aggregateSourceElements = new Map();
        typeNamePairs.forEach((entry) => {
            if (entry.name.includes('*')) {
                let keyMetadataType = entry.type;
                const metadataType = metadataTypeFactory_1.MetadataTypeFactory.getMetadataTypeFromMetadataName(keyMetadataType, this.swa.metadataRegistry);
                if (!metadataType) {
                    throw core_1.SfdxError.create('salesforce-alm', 'source', 'UnsupportedType', [keyMetadataType]);
                }
                if (metadataType.getMetadataName() !== metadataType.getAggregateMetadataName()) {
                    // In this case we are dealing with a decomposed item, so reload using the parent name
                    keyMetadataType = metadataType.getAggregateMetadataName();
                }
                const wildcards = [...sourceElements.keys()].filter((item) => item.includes(keyMetadataType))
                    .map((typeAnyName) => {
                    let [mdType, ...rest] = typeAnyName.split('__');
                    const mdName = rest.join('__');
                    return {
                        type: mdType,
                        name: mdName
                    };
                });
                aggregateSourceElements = new Map([...aggregateSourceElements, ...this._parseTypeNamePairs(wildcards, sourceElements)]);
            }
            else {
                const ase = SourceUtil.loadSourceElement(sourceElements, `${entry.type}__${entry.name}`, this.swa.metadataRegistry);
                aggregateSourceElements.set(ase.getKey(), ase);
            }
        });
        return aggregateSourceElements;
    }
    ;
    _processResults(result, aggregateSourceElements, deployDir) {
        return __awaiter(this, void 0, void 0, function* () {
            if (result.success && result.details.componentFailures) {
                this.removeFailedAggregates(result.details.componentFailures, aggregateSourceElements);
            }
            // We need to check both success and status because a status of 'SucceededPartial' returns success === true even though rollbackOnError is set.
            if (result.success && result.status === 'Succeeded') {
                const isNonDestructiveChangeDelete = this.isDelete && !fsExtra.existsSync(`${deployDir}${path.sep}destructiveChangesPost.xml`);
                result.outboundFiles = this.getOutboundFiles(aggregateSourceElements, isNonDestructiveChangeDelete);
                return result;
            }
            else {
                // throw the error that is created by _setupDeployFail
                throw yield this._setupDeployFail(result, aggregateSourceElements);
            }
        });
    }
    ;
    _setupDeployFail(result, aggSourceElements) {
        return __awaiter(this, void 0, void 0, function* () {
            const deployFailed = new Error();
            if (result.timedOut) {
                deployFailed.name = 'PollingTimeout';
            }
            else {
                deployFailed.name = 'DeployFailed';
                deployFailed.failures = syncCommandHelper.getDeployFailures(result, aggSourceElements, this.swa.metadataRegistry);
            }
            if (result.success && result.status === 'SucceededPartial') {
                deployFailed.outboundFiles = this.getOutboundFiles(aggSourceElements);
            }
            if (this.isDelete) {
                yield this._revertDeletions(aggSourceElements);
                const messages = core_1.Messages.loadMessages('salesforce-alm', 'source_delete');
                deployFailed.message = messages.getMessage('sourceDeleteFailure');
            }
            return deployFailed;
        });
    }
    ;
    // Revert all deletions since something went wrong and they were not deleted server side.
    // This copies all the files from the temporary location back to their original location.
    _revertDeletions(ases) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const ase of ases.values()) {
                const parentDir = path.dirname(ase.getMetadataFilePath());
                try {
                    yield core_1.fs.access(parentDir, core_1.fs.constants.R_OK);
                }
                catch (e) {
                    // If the parent directory does not exist, re-create it
                    yield core_1.fs.mkdirp(parentDir, core_1.fs.DEFAULT_USER_DIR_MODE);
                }
                // Re-create each workspace element
                for (const we of ase.getWorkspaceElements()) {
                    const backupPath = path.join(this.tmpBackupDeletions, path.basename(we.getSourcePath()));
                    fsExtra.copySync(backupPath, we.getSourcePath());
                }
            }
        });
    }
    ;
}
exports.SourceDeployApi = SourceDeployApi;

//# sourceMappingURL=sourceDeployApi.js.map
