"use strict";
/*
 * Copyright (c) 2018, Salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
Object.defineProperty(exports, "__esModule", { value: true });
const path = require('path');
const BBPromise = require('bluebird');
const replace = BBPromise.promisify(require('replace'));
const MdapiPollIntervalStrategy = require(path.join(__dirname, '..', 'mdapiPollIntervalStrategy'));
const MdapiDeployApi = require(path.join(__dirname, '..', 'mdapiDeployApi'));
const SourceConvertApi = require(path.join(__dirname, '..', 'sourceConvertApi'));
const { sequentialExecute } = require(path.join(__dirname, '..', 'srcDevUtil'));
const sourceState = require(path.join(__dirname, '..', 'sourceState'));
const sourceUtil_1 = require("./sourceUtil");
const consts = require(path.join(__dirname, '..', 'constants'));
const { INSTANCE_URL_TOKEN } = consts;
/**
 * Methods used to deploy source into an org and handle resulting output
 * Used by source:deploy and source:push commands
 */
class SourceDeployApiBase {
    constructor(orgApi) {
        this.orgApi = orgApi;
        this.force = orgApi.force;
    }
    /**
     * Set additional required MDAPI options config and deploy
     * @param options The flags on the context passed to the command.
     * @param {any} pollIntervalStrategy The strategy for how often to poll when doing a MD deploy
     * @returns {any}
     */
    mdapiDeploy(options, pollIntervalStrategy) {
        options.disableLogging = true;
        options.autoUpdatePackage = false;
        options.rollbackonerror = true;
        options.testlevel = options.testlevel || 'NoTestRun';
        return new MdapiDeployApi(this.orgApi, pollIntervalStrategy).deploy(options);
    }
    /**
     * Set the state of the source elements that were deployed
     * @param aggregateSourceElements The map of source elements to deploy
     * @returns {any}
     */
    getOutboundFiles(aggregateSourceElements, isDelete = false) {
        let deployedSourceElements = [];
        aggregateSourceElements.forEach((aggregateSourceElement) => {
            deployedSourceElements = deployedSourceElements.concat(isDelete ?
                aggregateSourceElement.getWorkspaceElements().filter(el => el.getState() === sourceState.DELETED) :
                aggregateSourceElement.getWorkspaceElements());
        });
        return deployedSourceElements.map(workspaceElement => workspaceElement.toObject());
    }
    /**
     * Convert the SDFX-formatted source to MD-format and then deploy it to the org
     * @param options The flags on the context passed to the command.
     * @param sourceWorkspaceAdapter
     * @param aggregateSourceElements The map of source elements to deploy
     * @param {boolean} createDestructiveChanges Whether destructiveChanges.xml needs to be created for this deployment
     * @returns {any}
     */
    convertAndDeploy(options, sourceWorkspaceAdapter, aggregateSourceElements, createDestructiveChanges) {
        const sourceConvertApi = new SourceConvertApi(this.orgApi, sourceWorkspaceAdapter);
        return sourceConvertApi.convertSourceToMdapi(options.deploydir, null, aggregateSourceElements, createDestructiveChanges, options.unsupportedMimeTypes, options.delete)
            .then(([sourceElementsToUpsert, deletedTypeNamePairs]) => {
            // replace tokens embedded in source
            if (options.replacetokens) {
                return this.replaceTokens(options.deploydir)
                    .then(() => [sourceElementsToUpsert, deletedTypeNamePairs]);
            }
            return [sourceElementsToUpsert, deletedTypeNamePairs];
        })
            .then(([sourceElementsToUpsert, deletedTypeNamePairs]) => {
            const pollIntervalStrategy = new MdapiPollIntervalStrategy(sourceElementsToUpsert, deletedTypeNamePairs);
            return this.mdapiDeploy(options, pollIntervalStrategy);
        })
            .catch(err => {
            if (err.name === 'mdapiDeployFailed') {
                return err.result;
            }
            throw err;
        });
    }
    /**
     * Remove source elements that failed to deploy
     * @param componentFailures
     * @param aggregateSourceElements
     */
    removeFailedAggregates(componentFailures, deployedSourceElements) {
        let failedSourcePaths = [];
        const failures = sourceUtil_1.toArray(componentFailures);
        failures.map((failure) => `${failure.componentType}__${failure.fullName}`).forEach((key) => {
            if (deployedSourceElements.has(key)) {
                deployedSourceElements.get(key).getWorkspaceElements().forEach((element) => {
                    failedSourcePaths = failedSourcePaths.concat(element.getSourcePath());
                });
                deployedSourceElements.delete(key);
            }
        });
    }
    /**
     * Some metadata types, such as RemoteSiteSetting, contain values specific to an org and need to be replaced
     * when the source is deployed to a new org
     * @param dir The directory where the metadata resides
     */
    replaceTokens(dir) {
        const replaceConfigs = [
            {
                regex: INSTANCE_URL_TOKEN,
                replacement: this.orgApi.authConfig.instanceUrl,
                paths: [dir],
                recursive: true,
                includes: ['*.remoteSite-meta.xml'],
                silent: true
            }
        ];
        const replaceFns = [];
        replaceConfigs.forEach((replaceConfig) => {
            replaceFns.push(() => new BBPromise((resolve, reject) => {
                try {
                    replace(replaceConfig);
                    return resolve();
                }
                catch (err) {
                    return reject(err);
                }
            }));
        });
        return sequentialExecute(replaceFns);
    }
}
exports.SourceDeployApiBase = SourceDeployApiBase;

//# sourceMappingURL=sourceDeployApiBase.js.map
