/*
 * Copyright (c) 2018, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = require("lodash");
const MetadataRegistry = require("../metadataRegistry");
const SourceWorkspaceAdapter = require("../sourceWorkspaceAdapter");
const SourceUtil = require("./sourceUtil");
const mdapiRetrieveApi_1 = require("../mdapiRetrieveApi");
const MdapiPullApi = require("../mdapiPullApi");
const core_1 = require("@salesforce/core");
const aggregateSourceElement_1 = require("../aggregateSourceElement");
const syncCommandHelper = require("../syncCommandHelper");
const metadataTypeFactory_1 = require("../metadataTypeFactory");
core_1.Messages.importMessagesDirectory(__dirname);
/**
 * API for retrieving metadata from an org and updating a local SFDX project.
 */
class SourceRetrieve {
    constructor(org) {
        this.org = org;
        this.projectPath = org.force.config.getProjectPath();
        this.defaultPackagePath = org.config.getAppConfig().defaultPackagePath;
    }
    // Retrieves metadata from an org based on the retrieve options specified.
    retrieve(options) {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger = yield core_1.Logger.child('SourceRetrieve');
            yield MetadataRegistry.initializeMetadataTypeInfos(this.org);
            // Only put SWA in stateless mode when sourcepath param is used.
            const mode = options.sourcepath && SourceWorkspaceAdapter.modes.STATELESS;
            this.swa = new SourceWorkspaceAdapter(this.org, this.defaultPackagePath, MetadataRegistry, true, mode);
            let results;
            try {
                this.tmpOutputDir = yield SourceUtil.createOutputDir('sourceRetrieve');
                if (options.sourcepath) {
                    this.logger.info(`Retrieving metadata in sourcepath '${options.sourcepath}' from org: '${this.org.name}'`);
                    results = yield this.retrieveFromSourcePath(options);
                }
                else if (options.manifest) {
                    this.logger.info(`Retrieving metadata in manifest '${options.manifest}' from org: '${this.org.name}'`);
                    results = yield this.retrieveFromManifest(options);
                }
                else if (options.metadata) {
                    this.logger.info(`Retrieving metadata '${options.metadata}' from org: '${this.org.name}'`);
                    results = yield this.retrieveFromMetadata(options);
                }
            }
            finally {
                yield SourceUtil.cleanupOutputDir(this.tmpOutputDir);
                // Delete the sourcePathInfos.json for this org.  Ideally, we never create it but
                // until then, this is necessary.
                this.org.getSourcePathInfos().delete();
            }
            return results;
        });
    }
    // Retrieve specific source paths from an org and update the project.
    retrieveFromSourcePath(options) {
        return __awaiter(this, void 0, void 0, function* () {
            // Parse the sourcepath parameter for metadata files and build a map of AggregateSourceElements
            const aggregateSourceElements = yield SourceUtil.getSourceElementsFromSourcePath(options.sourcepath, this.swa);
            // Convert aggregateSourceElements to an array of this format: { type: 'ApexClass', name: 'MyClass' }
            // for use by ManifestApi.createManifest().
            const mdFullPairs = [...aggregateSourceElements.values()].map(el => ({ type: el.getMetadataName(), name: el.getAggregateFullName() }));
            // Create a manifest and update the options with the manifest file.
            options.manifest = (yield SourceUtil.createManifest(this.org, options, mdFullPairs, this.tmpOutputDir)).file;
            // Now that we have a package.xml, the rest is just a retrieve using the manifest.
            return this.retrieveFromManifest(options, aggregateSourceElements);
        });
    }
    retrieveFromMetadata(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const config = yield this.org.getConfig();
            const connection = yield this.org.force._getConnection(this.org, this.org.config, config);
            const entries = yield SourceUtil.parseManifestEntries(options.metadata, connection, this.swa.metadataRegistry);
            // toManifest will also tack the manifest onto the options
            yield SourceUtil.toManifest(this.org, this.swa.metadataRegistry, options, this.tmpOutputDir);
            // Now that we have a package.xml, the rest is just a retrieve using the manifest.
            return this.retrieveFromManifest(options, null, entries);
        });
    }
    // Retrieve metadata specified in a manifest file (package.xml) from an org and update the project.
    retrieveFromManifest(options, aggregateSourceElements, entries) {
        return __awaiter(this, void 0, void 0, function* () {
            let results = { inboundFiles: [] };
            // Set Retrieve options for MdRetrieveApi
            const retrieveOptions = Object.assign(mdapiRetrieveApi_1.MdRetrieveApi.getDefaultOptions(), {
                retrievetargetdir: this.tmpOutputDir,
                unpackaged: options.manifest,
                wait: options.wait
            });
            // Retrieve the files from the org
            const res = yield new mdapiRetrieveApi_1.MdRetrieveApi(this.org).retrieve(retrieveOptions);
            // Convert to source format and update the local project.
            if (SourceRetrieve.isRetrieveSuccessful(res)) {
                const mdapiPull = new MdapiPullApi(this.org, this.swa);
                // Extracts the source from package.xml in the temp dir, creates a Map of AggregateSourceElements
                // and updates the local project with the files from the temp dir.
                const sourceElements = yield mdapiPull._syncDownSource(res, retrieveOptions, this.swa);
                let _entries = entries || (yield SourceUtil.parseToManifestEntriesArray(retrieveOptions.unpackaged));
                // Build a simple object representation of what was changed in the local project.
                results = this.processResults(res, sourceElements, aggregateSourceElements, _entries);
            }
            else {
                // If fileProperties is an object, it means no results were retrieved so don't throw an error
                if (!lodash_1.isPlainObject(res.fileProperties)) {
                    let errMsgExtra = '';
                    if (res.messages) {
                        errMsgExtra = `\nDue to: ${res.messages.problem}`;
                    }
                    throw core_1.SfdxError.create('salesforce-alm', 'source_retrieve', 'SourceRetrieveError', [errMsgExtra]);
                }
            }
            return results;
        });
    }
    validateAggregateFullName(source, target, delimiter) {
        if (target.type === source.type) {
            const fullNameComponents = source.name.split(delimiter);
            return fullNameComponents && fullNameComponents.length > 0 &&
                fullNameComponents[0] === target.name;
        }
    }
    processResults(result, sourceElements, aggregateSourceElements, entries) {
        const inboundFiles = [];
        /**
         * @todo We probably don't need both aggregateSourceElements and entries for filtering. We probably can rely on
         * what can be distilled from the manifest for filtering. In this case aggregateSourceElements is empty in the
         * case of the metadata scope. However for all three scope options we have a manifest to create workspace element
         * filtering.
         */
        const _aggregateSourceElements = aggregateSourceElements || this.swa.getAggregateSourceElements(false);
        // For each source element extracted from the zip (i.e., from the org) match it to entries in
        // the full map of AggregateSourceElements and create display rows for command output.
        sourceElements.forEach(sourceElement => {
            const key = aggregateSourceElement_1.AggregateSourceElement.getKeyFromMetadataNameAndFullName(sourceElement.getMetadataName(), sourceElement.getAggregateFullName());
            const se = _aggregateSourceElements.get(key);
            let filteredElements = se.getWorkspaceElements();
            if (!aggregateSourceElements) {
                /**
                 * This tell us that if true the sourceElement matched a wildcard in the entries. Therefore all child
                 * elements are not only included in the package zip but also should be included in the display.
                 */
                const matchesWildCardSourceElement = !!entries.find((element) => {
                    return sourceElement.getMetadataName() === element.type && element.name === '*';
                });
                // Remove all the workspace elements that don't match the scope. why?
                // If you don't and you do something like ListView:Foo__c.Foo You'll get all the members
                // of the custom object currently on disk. But what you really want is just a ListView.
                filteredElements = filteredElements.filter((workspaceElement) => {
                    const type = metadataTypeFactory_1.MetadataTypeFactory.getMetadataTypeFromMetadataName(workspaceElement.getMetadataName(), this.swa.metadataRegistry);
                    const parent = lodash_1.get(type, 'typeDefObj.parent');
                    // Find a manifest entry that matches the source element itself or the elements parent.
                    // If we wanted an entire CustomObject object we want to ensure all the child components are not
                    // filtered out.
                    //
                    // Anything not found in the entries array is filtered out.
                    return !!entries.find((manifestEntry) => {
                        if (parent && (parent.metadataName === manifestEntry.type)) {
                            return this.validateAggregateFullName({ name: workspaceElement.getFullName(), type: parent.metadataName }, manifestEntry, '.');
                        }
                        if (manifestEntry.type === workspaceElement.getMetadataName()) {
                            if (this.validateAggregateFullName({ name: workspaceElement.getFullName(), type: workspaceElement.getMetadataName() }, manifestEntry, '/')) {
                                /**
                                 * Case in point (-m AuraDefinitionBundle:xray )
                                 * The manifest entries will contain {name='xray', type='AuraDefinitionBundle'}
                                 * All the elements of the AuraDefinitionBundle will begin with the fullname 'xray'.
                                 * i.e. xray/xray.css. If that's the case return true otherwise continue evaluating.
                                 */
                                return true;
                            }
                        }
                        return matchesWildCardSourceElement || (manifestEntry.type === workspaceElement.getMetadataName() &&
                            manifestEntry.name === workspaceElement.getFullName());
                    });
                });
            }
            // Get all WorkspaceElements for the retrieved source elements and create simple object
            // data for display within a table.
            filteredElements.forEach(workspaceElement => {
                syncCommandHelper.createDisplayRows(inboundFiles, workspaceElement.toObject(), this.projectPath);
            });
        });
        const output = { inboundFiles };
        // If there are warning messages along with successes, display them (i.e., partial success).
        if (result.messages) {
            output.warnings = SourceUtil.toArray(result.messages);
        }
        return output;
    }
    static isRetrieveSuccessful(result) {
        return result &&
            result.success &&
            lodash_1.get(result, 'status') === 'Succeeded' &&
            Array.isArray(result.fileProperties);
    }
}
exports.SourceRetrieve = SourceRetrieve;

//# sourceMappingURL=sourceRetrieve.js.map
