/*
 * Copyright (c) 2018, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const core_1 = require("@salesforce/core");
const sourceRetrieve_1 = require("./sourceRetrieve");
const sourceUtil_1 = require("./sourceUtil");
const syncCommandHelper = require("../syncCommandHelper");
const logApi = require("../logApi");
core_1.Messages.importMessagesDirectory(__dirname);
/**
 * Command for retrieving metadata from a non-source-tracked org and updating a local SFDX project.
 */
class SourceRetrieveCommand {
    constructor() {
        this.logger = logApi.child('source:retrieve');
    }
    /**
     * Executes the source retrieve command
     * @param context - the cli context
     * @returns {Promise}
     */
    execute(context) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield new sourceRetrieve_1.SourceRetrieve(context.org).retrieve(context.flags);
        });
    }
    /**
     * Validates the source retrieve command parameters
     * @param context - the cli context
     * @returns {Promise}
     */
    validate(context) {
        return __awaiter(this, void 0, void 0, function* () {
            // Validate the wait param if set and convert to an integer.
            sourceUtil_1.parseWaitParam(context.flags);
            // verify that exactly one scoping flag is provided
            sourceUtil_1.validateXorFlags(context.flags);
            // verify that the org is not a scratch org
            yield sourceUtil_1.validateNonSourceTrackedOrg(context.org.name, 'pull');
            // verify that the manifest file exists and is readable.
            if (context.flags.manifest) {
                yield sourceUtil_1.validateManifestPath(context.flags.manifest);
            }
            return Promise.resolve(context);
        });
    }
    getHumanSuccessMessage(results) {
        const commonMsgs = core_1.Messages.loadMessages('salesforce-alm', 'source');
        const messages = core_1.Messages.loadMessages('salesforce-alm', 'source_retrieve');
        this.logger.styledHeader(this.logger.color.blue(messages.getMessage('retrievedSourceHeader')));
        if (results.inboundFiles && results.inboundFiles.length) {
            const columns = syncCommandHelper.getColumnMetaInfo(commonMsgs, true);
            this.logger.table(results.inboundFiles, { columns });
        }
        else {
            this.logger.log(messages.getMessage('NoResultsFound'));
        }
        if (results.warnings) {
            this.logger.log('');
            this.logger.styledHeader(this.logger.color.yellow(messages.getMessage('metadataNotFoundWarning')));
            results.warnings.forEach(warning => this.logger.log(warning.problem));
        }
    }
}
exports.SourceRetrieveCommand = SourceRetrieveCommand;

//# sourceMappingURL=sourceRetrieveCommand.js.map
