/*
 * Copyright (c) 2018, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = require('path');
const os = require("os");
const lodash_1 = require("lodash");
const aggregateSourceElement_1 = require("../aggregateSourceElement");
const metadataTypeFactory_1 = require("../metadataTypeFactory");
const core_1 = require("@salesforce/core");
const mdapiRetrieveApi_1 = require("../mdapiRetrieveApi");
const inFolderMetadataType_1 = require("../metadataTypeImpl/inFolderMetadataType");
const ManifestCreateApi = require("../manifestCreateApi");
const srcDevUtil = require(path.join(__dirname, '..', 'srcDevUtil'));
const constants = require(path.join(__dirname, '..', 'constants'));
const MdapiPackage = require(path.join(__dirname, '..', 'mdapiPackage'));
core_1.Messages.importMessagesDirectory(__dirname);
/**
 * Validate the value for the 'wait' parameter and reset it as a number.
 * @param flags The command parameters (aka flags)
 */
exports.parseWaitParam = (flags) => {
    const { wait } = flags;
    if (!lodash_1.isNil(wait)) {
        if (srcDevUtil.isInt(wait)) {
            flags.wait = parseInt(wait, 10); // convert to a number
            if (flags.wait >= constants.MIN_SRC_WAIT_MINUTES) {
                return;
            }
        }
        const errConfig = new core_1.SfdxErrorConfig('salesforce-alm', 'source', 'mdapiCliInvalidNumericParam');
        errConfig.setErrorTokens(['wait']);
        throw core_1.SfdxError.create(errConfig);
    }
};
/**
 * Validate that 1 and only 1 scoping flag was specified for the command.
 * @param flags The command parameters (aka flags)
 */
exports.validateXorFlags = (flags) => {
    const xorFlags = ['manifest', 'metadata', 'sourcepath'];
    const xorFlagCount = Object.keys(flags).reduce((sum, key) => xorFlags.includes(key) ? sum + 1 : sum, 0);
    // verify that exactly one scoping flag is provided
    if (xorFlagCount !== 1) {
        const errConfig = new core_1.SfdxErrorConfig('salesforce-alm', 'source', 'OneScopingParamError');
        errConfig.addAction('OneScopingParamErrorAction');
        throw core_1.SfdxError.create(errConfig);
    }
};
/**
 * Validate that the org is a non-source-tracked org.
 * @param orgName The username of the org for doing the source:deploy or source:retrieve
 * @param errAction The action ('push' or 'pull') to take when the org is discovered to be a source tracked org.
 */
exports.validateNonSourceTrackedOrg = (orgName, errAction) => __awaiter(this, void 0, void 0, function* () {
    if (yield srcDevUtil.isSourceTrackedOrg(orgName)) {
        const errConfig = new core_1.SfdxErrorConfig('salesforce-alm', 'source', 'SourceTrackedOrgError');
        errConfig.addAction('SourceTrackedOrgErrorAction', [errAction]);
        throw core_1.SfdxError.create(errConfig);
    }
});
/**
 * Validate that a manifest file path exists and is readable.
 * @param manifestPath The path to the manifest file (package.xml)
 */
exports.validateManifestPath = (manifestPath) => __awaiter(this, void 0, void 0, function* () {
    try {
        yield core_1.fs.access(manifestPath, core_1.fs.constants.R_OK);
    }
    catch (e) {
        throw core_1.SfdxError.create('salesforce-alm', 'source', 'InvalidManifestError', [manifestPath]);
    }
});
function createOutputDir(cmdName) {
    return __awaiter(this, void 0, void 0, function* () {
        const logger = yield core_1.Logger.child('SourceUtil');
        const targetDir = process.env.SFDX_SOURCE_TARGET_DIR || os.tmpdir();
        const tmpOutputDir = path.join(targetDir, `sdx_${cmdName}_${Date.now()}`);
        yield core_1.fs.mkdirp(tmpOutputDir, core_1.fs.DEFAULT_USER_DIR_MODE);
        logger.info(`Created output directory '${tmpOutputDir}'`);
        return tmpOutputDir;
    });
}
exports.createOutputDir = createOutputDir;
function cleanupOutputDir(outputDir) {
    return __awaiter(this, void 0, void 0, function* () {
        const logger = yield core_1.Logger.child('SourceUtil');
        if (outputDir && !outputDir.includes(process.env.SFDX_SOURCE_TARGET_DIR)) {
            try {
                yield core_1.fs.remove(outputDir);
                if (yield core_1.fs.stat(`${outputDir}.zip`)) {
                    yield core_1.fs.remove(`${outputDir}.zip`);
                }
            }
            catch (err) {
                logger.warn(`Could not delete the outputDir '${outputDir}' due to: ${err.message}`);
            }
        }
        else {
            logger.warn(`Did not delete the outputDir '${outputDir}' because it was set by the user`);
        }
    });
}
exports.cleanupOutputDir = cleanupOutputDir;
/**
 * Return the aggregate source element for the specified file
 * @param {string} sourcePath the file in the workspace
 * @param sourceWorkspaceAdapter
 * @returns {AggregateSourceElement}
 */
exports.getSourceElementForFile = function (sourcePath, sourceWorkspaceAdapter, metadataType) {
    let aggregateSourceElement;
    const mdRegistry = sourceWorkspaceAdapter.metadataRegistry;
    const sourceElementMetadataType = metadataType || metadataTypeFactory_1.MetadataTypeFactory.getMetadataTypeFromSourcePath(sourcePath, mdRegistry);
    if (sourceElementMetadataType) {
        // This will build an AggregateSourceElement with only the specified WorkspaceElement
        // (child element) when the metadata type has a parent.
        const _sourcePath = lodash_1.get(sourceElementMetadataType, 'typeDefObj.parent') ? sourcePath : path.dirname(sourcePath);
        const aggregateMetadataName = sourceElementMetadataType.getAggregateMetadataName();
        const aggregateFullName = sourceElementMetadataType.getAggregateFullNameFromFilePath(sourcePath);
        const key = aggregateSourceElement_1.AggregateSourceElement.getKeyFromMetadataNameAndFullName(aggregateMetadataName, aggregateFullName);
        // Get the AggregateSourceElement, which will only populate with the specified WorkspaceElement
        // when sourcePath is part of an ASE.
        const sourceElements = sourceWorkspaceAdapter.getAggregateSourceElements(false, undefined, undefined, _sourcePath);
        aggregateSourceElement = exports.loadSourceElement(sourceElements, key, mdRegistry);
    }
    else {
        throw core_1.SfdxError.create('salesforce-alm', 'source', 'SourcePathInvalid', [sourcePath]);
    }
    return aggregateSourceElement;
};
/**
 * Get the source elements from the source path, whether for a particular file or a directory
 * @param {string} optionsSourcePath
 * @param {any} sourceWorkspaceAdapter
 * @returns {Map<string, AggregateSourceElement>}
 */
exports.getSourceElementsFromSourcePath = function (optionsSourcePath, sourceWorkspaceAdapter) {
    return __awaiter(this, void 0, void 0, function* () {
        let aggregateSourceElements = new Map();
        const mdRegistry = sourceWorkspaceAdapter.metadataRegistry;
        for (let sourcepath of optionsSourcePath.split(',')) {
            // resolve to an absolute path
            sourcepath = path.resolve(sourcepath.trim());
            // Throw an error if the source path isn't accessible.
            try {
                yield core_1.fs.access(sourcepath, core_1.fs.constants.R_OK);
            }
            catch (e) {
                throw core_1.SfdxError.create('salesforce-alm', 'source', 'SourcePathInvalid', [sourcepath]);
            }
            // Get the MetadataType so we can resolve the path.  Some paths such as individual static
            // resources need to use a different path when getting source elements.
            const metadataType = metadataTypeFactory_1.MetadataTypeFactory.getMetadataTypeFromSourcePath(sourcepath, mdRegistry);
            if (metadataType) {
                sourcepath = metadataType.resolveSourcePath(sourcepath);
            }
            // Get a single source element or a directory of source elements and add it to the map.
            if (srcDevUtil.containsFileExt(sourcepath)) {
                const ase = exports.getSourceElementForFile(sourcepath, sourceWorkspaceAdapter, metadataType);
                const aseKey = ase.getKey();
                if (aggregateSourceElements.has(aseKey)) {
                    // We already have one of the child elements of this ASE so update the current ASE
                    const _ase = aggregateSourceElements.get(aseKey);
                    _ase.addWorkspaceElement(ase.getWorkspaceElements()[0]);
                    aggregateSourceElements.set(aseKey, _ase);
                }
                else {
                    aggregateSourceElements.set(aseKey, ase);
                }
            }
            else {
                aggregateSourceElements = new Map([...aggregateSourceElements, ...exports.getSourceElementsInPath(sourcepath, sourceWorkspaceAdapter)]);
            }
        }
        ;
        return aggregateSourceElements;
    });
};
/**
 * Return the specified aggregate source element or error if it does not exist
 * @param {Map<string, AggregateSourceElement>} sourceElements All the source elements in the workspace
 * @param {string} key The key of the particular source element we are looking for
 * @param {any} metadataRegistry
 * @returns {AggregateSourceElement}
 */
exports.loadSourceElement = function (sourceElements, key, metadataRegistry) {
    let aggregateSourceElement = sourceElements.get(key);
    if (!aggregateSourceElement) {
        // Namespaces also contain '__' so only split on the first occurrence of '__'
        let [mdType, ...rest] = key.split('__');
        const mdName = rest.join('__');
        const metadataType = metadataTypeFactory_1.MetadataTypeFactory.getMetadataTypeFromMetadataName(mdType, metadataRegistry);
        const hasParentType = metadataType.getMetadataName() !== metadataType.getAggregateMetadataName();
        if (hasParentType) { //In this case, we are dealing with a decomposed subtype, so need to check for a parent
            const parentName = metadataType.getAggregateMetadataName();
            const parentMetadataType = metadataTypeFactory_1.MetadataTypeFactory.getAggregateMetadataType(parentName, metadataRegistry);
            const parentFullName = metadataType.getAggregateFullNameFromWorkspaceFullName(mdName);
            const newKey = aggregateSourceElement_1.AggregateSourceElement.getKeyFromMetadataNameAndFullName(parentMetadataType.getAggregateMetadataName(), parentFullName);
            return exports.loadSourceElement(sourceElements, newKey, metadataRegistry);
        }
        else if (metadataType instanceof inFolderMetadataType_1.InFolderMetadataType) {
            mdType = MdapiPackage.convertFolderTypeKey(mdType);
            return exports.loadSourceElement(sourceElements, `${mdType}__${mdName}`, metadataRegistry);
        }
        else {
            const errConfig = new core_1.SfdxErrorConfig('salesforce-alm', 'source_deploy', 'SourceElementDoesNotExist');
            errConfig.setErrorTokens([mdType, mdName]);
            throw core_1.SfdxError.create(errConfig);
        }
    }
    return aggregateSourceElement;
};
/**
 * Return the aggregate source elements found in the provided source path
 * @param {Array<string>} sourcePath The path to look for source elements in
 * @param sourceWorkspaceAdapter
 * @returns {Map<string, AggregateSourceElement>}
 */
exports.getSourceElementsInPath = function (sourcePath, sourceWorkspaceAdapter) {
    return sourceWorkspaceAdapter.getAggregateSourceElements(false, undefined, undefined, sourcePath);
};
/**
 * Convert the argument into an array datatype
 * @param arrayOrObjectOrUndefined
 * @returns Array
 */
exports.toArray = function (arrayOrObjectOrUndefined) {
    if (!arrayOrObjectOrUndefined) {
        return [];
    }
    else if (Array.isArray(arrayOrObjectOrUndefined)) {
        return arrayOrObjectOrUndefined;
    }
    else {
        return [arrayOrObjectOrUndefined];
    }
};
/**
 * Parse the manifest file and create a list ManifestEntry objects.
 * @param manifestPath {string} The filepath for the manifest
 * @returns {ManifestEntry[]} An array for ManifestEntry objects from the manifest.
 */
exports.parseToManifestEntriesArray = function (manifestPath) {
    return __awaiter(this, void 0, void 0, function* () {
        const entries = [];
        const options = {
            unpackaged: manifestPath
        };
        return mdapiRetrieveApi_1.MdRetrieveApi._getPackageJson(undefined, options)
            .then((manifestJson) => {
            exports.toArray(manifestJson.types).forEach((type) => {
                if (!type.name) {
                    const errConfig = new core_1.SfdxErrorConfig('salesforce-alm', 'source', 'IllFormattedManifest');
                    errConfig.setErrorTokens(['; <name> is missing']);
                    throw core_1.SfdxError.create(errConfig);
                }
                exports.toArray(type.members).forEach((member) => {
                    entries.push({
                        type: type.name,
                        name: member
                    });
                });
            });
            return entries;
        });
    });
};
/**
 * Parse manifest entry strings into an array of ManifestEntry objects
 * @param arg {string} The entry string "ApexClass: *, AuraDefinitionBundle
 * @param connection { any } The jsforce connection to the org.
 * @param metadataRegistry An instance of MetadataRegistry, typically from an instance of SourceWorkspaceAdapter
 */
exports.parseManifestEntries = function (arg, connection, metadataRegistry) {
    return __awaiter(this, void 0, void 0, function* () {
        if (arg) {
            const mdParamArray = arg.split(',');
            let manifestEntries = [];
            for (const md of mdParamArray) {
                const [mdType, ...rest] = md.split(':');
                const mdName = rest.length ? rest.join(':') : '*';
                const metadataType = metadataTypeFactory_1.MetadataTypeFactory.getMetadataTypeFromMetadataName(mdType, metadataRegistry);
                if (metadataType instanceof inFolderMetadataType_1.InFolderMetadataType && mdName === '*') {
                    // We need to make a few listMetdata() server calls to get folder and child metadata.
                    // Unfortunately, we have to wrap calls in toArray since the listMetadata() server API
                    // does not necessarily return an array...<sigh>.
                    const type = MdapiPackage.convertFolderTypeKey(mdType);
                    const folders = exports.toArray(yield connection.metadata.list([{ type }]));
                    for (const f of folders) {
                        manifestEntries.push({ type: mdType, name: f.fullName });
                        const folderChildren = exports.toArray(yield connection.metadata.list([{ type: mdType, folder: f.fullName }]));
                        for (const child of folderChildren) {
                            manifestEntries.push({ type: mdType, name: child.fullName });
                        }
                    }
                }
                else {
                    manifestEntries.push({ type: mdType.trim(), name: mdName.trim() });
                }
            }
            return manifestEntries;
        }
        return null;
    });
};
/**
 * Converts SourceOptions.metadata into a package manifest for a given org.
 * @param org { any } The org
 * @param metadataRegistry An instance of MetadataRegistry, typically from an instance of SourceWorkspaceAdapter
 * @param options { SourceOptions } The source options containing the metadata
 * @returns {Promise<string | null>} A path to the created manifest or null of options or options.metadata is null.
 */
exports.toManifest = function (org, metadataRegistry, options, tmpOutputDir) {
    return __awaiter(this, void 0, void 0, function* () {
        if (options && options.metadata) {
            const config = yield org.getConfig();
            const connection = yield org.force._getConnection(org, org.config, config);
            const entries = yield exports.parseManifestEntries(options.metadata, connection, metadataRegistry);
            if (entries != null) {
                // Create a manifest and update the options with the manifest file.
                options.manifest = (yield exports.createManifest(org, options, entries, tmpOutputDir)).file;
                return options.manifest;
            }
            else {
                return null;
            }
        }
        return null;
    });
};
/**
 * Function to create a manifest for a given org
 * @param org {AnyJson} An org
 * @param options {SourceOptions} Source options
 * @param mdPairs {ManifestEntry[]} Array of metadata items
 * @returns A package.xml manifest
 */
exports.createManifest = function (org, options, mdPairs = [], tmpOutputDir) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!org || !options) {
            return null;
        }
        const manifestApi = new ManifestCreateApi(org);
        // Create the package.xml in the temp dir
        const manifestOptions = Object.assign({}, options, {
            outputdir: tmpOutputDir
        });
        return manifestApi.createManifest(manifestOptions, null, mdPairs);
    });
};

//# sourceMappingURL=sourceUtil.js.map
