/*
 * Copyright, 1999-2016, salesforce.com
 * All Rights Reserved
 * Company Confidential
 */
'use strict';
const util = require('util');
const path = require('path');
const sourceState = require(path.join(__dirname, 'sourceState'));
const logger = require(path.join(__dirname, 'logApi'));
const AggregateSourceElement = require(path.join(__dirname, 'aggregateSourceElement')).AggregateSourceElement;
const WorkspaceElement = require(path.join(__dirname, 'workspaceElement')).WorkspaceElement;
const MetadataTypeFactory = require(path.join(__dirname, 'metadataTypeFactory')).MetadataTypeFactory;
const _getSourceLocationsKey = function (metadataName, aggregateFullName) {
    return `${metadataName}__${aggregateFullName}`;
};
/**
 * Utility to convert between metadata fullnames and workspace paths
 * @param sourceFileInfos - Source elements to manage
 * @private
 * @constructor
 */
function _SourceLocations(sourceFileInfos, metadataRegistry) {
    this.typeAndFullNameToMetadataPath = new Map();
    for (const sourceFileInfo of sourceFileInfos) {
        if (sourceFileInfo.isMetadataFile) {
            const pathMetadataType = MetadataTypeFactory.getMetadataTypeFromSourcePath(sourceFileInfo.sourcePath, metadataRegistry);
            if (pathMetadataType) {
                const aggregateFullName = pathMetadataType.getAggregateFullNameFromFilePath(sourceFileInfo.sourcePath);
                const sourceLocationsKey = _getSourceLocationsKey(pathMetadataType.getMetadataName(), aggregateFullName);
                const aggregateMetadataPath = pathMetadataType.getAggregateMetadataFilePathFromWorkspacePath(sourceFileInfo.sourcePath);
                this.typeAndFullNameToMetadataPath.set(sourceLocationsKey, aggregateMetadataPath);
            }
        }
    }
}
/**
 * Adapter between scratch org source metadata and a local workspace
 * @param {String} org - The org that the source workspace files belong to
 * @param {String} defaultPackagePath - The name of the default package path to which new source will be added
 * @constructor
 */
function SourceWorkspaceAdapter(org, defaultPackagePath, MetadataRegistry, fromConvert, sourceMode) {
    this.logger = logger.child('sourceWorkspaceAdapter');
    const SourcePathStatusManager = require(path.join(__dirname, 'sourcePathStatusManager')); // eslint-disable-line global-require
    this.isStateless = sourceMode === SourceWorkspaceAdapter.modes.STATELESS;
    this.spsm = new SourcePathStatusManager(org, this.isStateless);
    this.metadataRegistry = new MetadataRegistry(org);
    this.sourceLocations = new _SourceLocations(this.spsm.workspacePathInfos.values(), this.metadataRegistry);
    this.wsPath = org.config.getProjectPath();
    this.namespace = org.config.getAppConfig().namespace;
    this.defaultSrcDir = path.join(this.wsPath, defaultPackagePath, 'main', 'default');
    this.fromConvert = fromConvert || false;
    this.forceIgnore = this.spsm.forceIgnore;
    this.defaultPackagePath = org.config.getAppConfig().defaultPackagePath;
    this.pendingSourcePathInfos = new Map();
    // Array of sourcePathInfos for directories
    this.pendingDirectories = [];
    // mdapi:convert and source:convert do not need to load changed aggregate source elements.
    // The cache is specific to push, pull, status commands, but sourceWorkspaceAdapter is
    // initialized for all source-related commands
    if (!this.fromConvert) {
        this.changedSourceElementsCache = this.getAggregateSourceElements(true, null, true);
    }
}
SourceWorkspaceAdapter.modes = { STATE: 0, STATELESS: 1 };
SourceWorkspaceAdapter.prototype.revertSourcePathInfos = function () {
    this.spsm.revert();
};
SourceWorkspaceAdapter.prototype.backupSourcePathInfos = function () {
    this.spsm.backup();
};
/**
 * private helper to log when a metadata type isn't supported
 * @param {string} metadataName - the metadata type name
 * @param {string} filePath - the filepath to the metadata item
 * @private
 */
const _logUnsupported = function (metadataName, filePath) {
    if (!util.isNullOrUndefined(filePath)) {
        this.logger.warn(`Unsupported source member ${metadataName} at ${filePath}`);
    }
    else {
        this.logger.warn(`Unsupported source member ${metadataName}`);
    }
};
/**
 * Get AggregateSourceElements (ASEs) in the workspace.
 *
 * To get all ASEs: SourceWorkspaceAdapter.getAggregateSourceElements(false);
 *    NOTE: This caches all ASEs so that subsequent calls do not incur this perf hit and just return the cache.
 *
 * To get all changed ASEs: SourceWorkspaceAdapter.getAggregateSourceElements(true);
 *
 * To get only ASEs from a certain path: SourceWorkspaceAdapter.getAggregateSourceElements(false, undefined, undefined, myPath);
 *
 * @param {boolean} changesOnly - If true then return only the updated source elements (changed, new, or deleted)
 * @param {string} packageDirectory - the package directory from which to fetch source
 * @param {boolean} updatePendingPathInfos - the pending path infos should only be updated the first time this method is called
 *      in order to prevent subsequent calls from overwriting its values
 * @param {string} sourcePath the directory or file path specified for change-set development
 * @returns {Map<String, AggregateSourceElement>} - Map of aggregate source element key to aggregateSourceElement
 * ex. { 'ApexClass__myApexClass' : aggregateSourceElement }
 */
SourceWorkspaceAdapter.prototype.getAggregateSourceElements = function (changesOnly, packageDirectory, updatePendingPathInfos = false, sourcePath) {
    if (!changesOnly && this.allAggregateSourceElementsCache) {
        return this.allAggregateSourceElementsCache;
    }
    const aggregateSourceElements = new Map();
    // Retrieve sourcePathInfos from the manager, filtered by what we want.
    const pendingChanges = this.spsm.getSourcePathInfos({ changesOnly, packageDirectory, sourcePath });
    pendingChanges.forEach((change) => {
        // Skip the directories
        if (change.isDirectory) {
            if (updatePendingPathInfos) {
                this.pendingDirectories.push(change);
            }
            return;
        }
        const workspaceElementMetadataType = MetadataTypeFactory.getMetadataTypeFromSourcePath(change.sourcePath, this.metadataRegistry);
        // If this isn't a source file skip it
        if (!workspaceElementMetadataType) {
            return;
        }
        // Does the metadata registry have this type blacklisted.
        if (!this.metadataRegistry.isSupported(workspaceElementMetadataType.getMetadataName())) {
            _logUnsupported.call(this, workspaceElementMetadataType.getMetadataName(), change.sourcePath);
            return;
        }
        const aggregateFullName = workspaceElementMetadataType.getAggregateFullNameFromFilePath(change.sourcePath);
        const aggregateMetadataFilePath = workspaceElementMetadataType.getAggregateMetadataFilePathFromWorkspacePath(change.sourcePath);
        const aggregateMetadataType = MetadataTypeFactory.getAggregateMetadataType(workspaceElementMetadataType.getAggregateMetadataName(), this.metadataRegistry);
        const newAggregateSourceElement = new AggregateSourceElement(aggregateMetadataType, aggregateFullName, aggregateMetadataFilePath, this.metadataRegistry);
        const workspaceFullName = workspaceElementMetadataType.getFullNameFromFilePath(change.sourcePath);
        const deleteSupported = workspaceElementMetadataType.deleteSupported(workspaceFullName);
        const workspaceElement = new WorkspaceElement(workspaceElementMetadataType.getMetadataName(), workspaceFullName, change.sourcePath, change.state, deleteSupported);
        const key = newAggregateSourceElement.getKey();
        let aggregateSourceElement;
        if (aggregateSourceElements.has(key)) {
            aggregateSourceElement = aggregateSourceElements.get(key);
        }
        else {
            aggregateSourceElement = newAggregateSourceElement;
        }
        aggregateSourceElement.addWorkspaceElement(workspaceElement);
        aggregateSourceElements.set(key, aggregateSourceElement);
        if (updatePendingPathInfos) {
            this.pendingSourcePathInfos.set(change.sourcePath, change);
        }
    });
    if (!changesOnly && !sourcePath) {
        // We just created ASEs for all workspace elements so cache it.
        this.allAggregateSourceElementsCache = aggregateSourceElements;
    }
    return aggregateSourceElements;
};
/**
 * Commit pending changed file infos
 * @returns {boolean} - Was the commit successful
 */
SourceWorkspaceAdapter.prototype.commitPendingChanges = function () {
    if (util.isNullOrUndefined(this.pendingSourcePathInfos)
        && util.isNullOrUndefined(this.pendingDirectories)) {
        // getChanges or getAll must have been called prior to find the list
        //   of pending changes
        return false;
    }
    let pendingChanges = [];
    if (!util.isNullOrUndefined(this.pendingDirectories)) { // be wary of directories that get deleted by cleanup methods
        pendingChanges = pendingChanges.concat(this.pendingDirectories);
    }
    this.pendingSourcePathInfos.forEach(pendingSourcePathInfo => {
        pendingChanges.push(pendingSourcePathInfo);
    });
    this.spsm.commitChangedPathInfos(pendingChanges);
    return true;
};
/**
 * Update the source stored in the workspace
 * @param sourceElements - Collection of source elements to update
 */
SourceWorkspaceAdapter.prototype.updateSource = function (sourceElements, force, sOrgApi, manifest, checkForDuplicates, unsupportedMimeTypes) {
    let updatedPaths = [];
    let deletedPaths = [];
    if (checkForDuplicates) {
        sourceElements.forEach(se => se.checkForDuplicates());
    }
    sourceElements.forEach(sourceElement => {
        const [newPathsForElements, updatedPathsForElements, deletedPathsForElements] = sourceElement.commit(manifest, unsupportedMimeTypes);
        updatedPaths = updatedPaths.concat(newPathsForElements, updatedPathsForElements);
        deletedPaths = deletedPaths.concat(deletedPathsForElements);
    });
    this.spsm.updateInfosForPaths(updatedPaths, deletedPaths);
    return sourceElements;
};
/**
 * Create a source element representation of a metadata change in the local workspace
 * @param fileProperty - The file property from the retrieve result
 * @returns {null} - A source element or null if metadataType is not supported
 */
SourceWorkspaceAdapter.prototype.processMdapiFileProperty = function (changedSourceElements, retrieveRoot, fileProperty, bundleFileProperties) {
    // Right now, all fileProperties returned by the mdapi are for aggregate metadata types
    const aggregateMetadataType = MetadataTypeFactory.getMetadataTypeFromFileProperty(fileProperty, this.metadataRegistry);
    const aggregateFullName = aggregateMetadataType.getAggregateFullNameFromFileProperty(fileProperty, this.namespace);
    let aggregateMetadataPath;
    if (!this.metadataRegistry.isSupported(aggregateMetadataType.getMetadataName())) {
        return null;
    }
    aggregateMetadataPath = this.sourceLocations.getMetadataPath(aggregateMetadataType.getMetadataName(), aggregateFullName);
    if (!util.isNullOrUndefined(aggregateMetadataPath) && this.fromConvert && !this.defaultSrcDir.includes(this.defaultPackagePath) && !aggregateMetadataPath.includes(this.defaultSrcDir)) {
        // if a user specified a destination folder outside the default package directory and
        // a file with same type and name exists but in a different pacakge directory then ignore it
        aggregateMetadataPath = null;
    }
    const workspaceElementsToDelete = aggregateMetadataType.getWorkspaceElementsToDelete(aggregateMetadataPath, fileProperty);
    // If the metadata path wasn't found we will use the default source directory
    if (util.isNullOrUndefined(aggregateMetadataPath)) {
        aggregateMetadataPath = aggregateMetadataType.getDefaultAggregateMetadataPath(aggregateFullName, this.defaultSrcDir, bundleFileProperties);
        // Add the new path to the location mapping
        this.sourceLocations.addMetadataPath(aggregateMetadataType.getAggregateMetadataName(), aggregateFullName, aggregateMetadataPath);
    }
    if (this.forceIgnore.accepts(aggregateMetadataPath)) {
        const newAggregateSourceElement = new AggregateSourceElement(aggregateMetadataType, aggregateFullName, aggregateMetadataPath, this.metadataRegistry);
        const key = AggregateSourceElement.getKeyFromMetadataNameAndFullName(aggregateMetadataType.getMetadataName(), aggregateFullName);
        let aggregateSourceElement;
        if (changedSourceElements.has(key)) {
            aggregateSourceElement = changedSourceElements.get(key);
        }
        else {
            aggregateSourceElement = newAggregateSourceElement;
        }
        if (workspaceElementsToDelete.length > 0) {
            workspaceElementsToDelete.forEach(deletedElement => {
                aggregateSourceElement.addPendingDeletedWorkspaceElement(deletedElement);
            });
        }
        aggregateSourceElement.retrievedMetadataPath = aggregateMetadataType.getRetrievedMetadataPath(fileProperty, retrieveRoot, bundleFileProperties);
        const retrievedContentPath = aggregateMetadataType.getRetrievedContentPath(fileProperty, retrieveRoot);
        if (retrievedContentPath) {
            if (!aggregateSourceElement.retrievedContentPaths) {
                aggregateSourceElement.retrievedContentPaths = [];
            }
            aggregateSourceElement.retrievedContentPaths.push(retrievedContentPath);
        }
        changedSourceElements.set(key, aggregateSourceElement);
        return aggregateSourceElement;
    }
    return null;
};
/**
 * Create a source element representation of a deleted metadata change in the local workspace
 * @returns {null} - A source element or null if metadataType is not supported
 */
SourceWorkspaceAdapter.prototype.handleObsoleteSource = function (changedSourceElements, fullName, type) {
    const sourceMemberMetadataType = MetadataTypeFactory.getMetadataTypeFromMetadataName(type, this.metadataRegistry);
    const aggregateFullName = sourceMemberMetadataType.getAggregateFullNameFromSourceMemberName(fullName);
    const metadataPath = this.sourceLocations.getMetadataPath(sourceMemberMetadataType.getAggregateMetadataName(), aggregateFullName);
    if (metadataPath !== undefined) {
        const key = AggregateSourceElement.getKeyFromMetadataNameAndFullName(sourceMemberMetadataType.getAggregateMetadataName(), aggregateFullName);
        let aggregateSourceElement;
        if (changedSourceElements.has(key)) {
            aggregateSourceElement = changedSourceElements.get(key);
        }
        else {
            const aggregateMetadataType = MetadataTypeFactory.getAggregateMetadataType(sourceMemberMetadataType.getMetadataName(), this.metadataRegistry); // only used in one place ( to create AggregateSourceElement )
            aggregateSourceElement = new AggregateSourceElement(aggregateMetadataType, aggregateFullName, metadataPath, this.metadataRegistry);
        }
        const shouldDeleteWorkspaceAggregate = sourceMemberMetadataType.shouldDeleteWorkspaceAggregate(type);
        if (shouldDeleteWorkspaceAggregate) {
            aggregateSourceElement.markForDelete();
        }
        else { // the type is decomposed or AuraDefinition and we only want to delete components of an aggregate element
            const sourcePathsToDelete = aggregateSourceElement.getWorkspacePathsForTypeAndFullName(sourceMemberMetadataType.getMetadataName(), fullName);
            sourcePathsToDelete.forEach(sourcePathToDelete => {
                const deletedWorkspaceElement = new WorkspaceElement(sourceMemberMetadataType.getMetadataName(), fullName, sourcePathToDelete, sourceState.DELETED, true);
                aggregateSourceElement.addPendingDeletedWorkspaceElement(deletedWorkspaceElement);
            });
        }
        changedSourceElements.set(key, aggregateSourceElement);
        return aggregateSourceElement;
    }
    return null;
};
/**
 * Get the metadata file path for a fullname
 * @param metadataType - The type of the source
 * @param fullName - The SFDC fullname
 * @returns {path} - The path to the metadata file for the source
 */
_SourceLocations.prototype.getMetadataPath = function (metadataType, fullName) {
    return this.typeAndFullNameToMetadataPath.get(_getSourceLocationsKey(metadataType, fullName));
};
/**
 * Add a metadata file path for a fullName
 * @param metadataType - The type of the source
 * @param fullName - The SFDC fullname
 * @param metadataPath - The path to the metadata file for the source
 */
_SourceLocations.prototype.addMetadataPath = function (metadataType, fullName, metadataPath) {
    this.typeAndFullNameToMetadataPath.set(_getSourceLocationsKey(metadataType, fullName), metadataPath);
};
module.exports = SourceWorkspaceAdapter;

//# sourceMappingURL=sourceWorkspaceAdapter.js.map
