/*
 * Copyright (c) 2016, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root  or https://opensource.org/licenses/BSD-3-Clause
 */
'use strict';
// 3pp
const Promise = require('bluebird');
// Node
const path = require('path');
const util = require('util');
// Local
const SourceWorkspaceAdapter = require(path.join(__dirname, 'sourceWorkspaceAdapter'));
const MetadataRegistry = require(path.join(__dirname, 'metadataRegistry'));
const srcDevUtil = require(path.join(__dirname, 'srcDevUtil.js'));
const sourceState = require(path.join(__dirname, 'sourceState.js'));
const AggregateSourceElement = require(path.join(__dirname, 'aggregateSourceElement')).AggregateSourceElement;
const MetadataTypeFactory = require(path.join(__dirname, 'metadataTypeFactory')).MetadataTypeFactory;
const ForceIgnore = require(path.join(__dirname, 'forceIgnore')).ForceIgnore;
class SrcStatusApi {
    constructor(scratchOrg, swa) {
        this.scratchOrg = scratchOrg;
        this.force = scratchOrg.force;
        const appConfig = this.force.getConfig().getAppConfig();
        this.swa = swa || new SourceWorkspaceAdapter(this.scratchOrg, appConfig.defaultPackagePath, MetadataRegistry);
        this.maxRevisionFile = this.scratchOrg.getMaxRevision();
        this.locallyChangedWorkspaceElements = [];
        this.localChanges = [];
        this.remoteChanges = [];
        this.forceIgnore = new ForceIgnore();
    }
    doStatus(options) {
        return Promise.resolve()
            .then(() => this._populateLocalChanges(options))
            .then(() => this._populateServerChanges(options))
            .then(() => this._markConflicts(options));
    }
    _populateLocalChanges(options) {
        if (!options.local) {
            return [];
        }
        const localSourceElementsMap = this.swa.changedSourceElementsCache;
        return localSourceElementsMap.forEach((value) => {
            value.getWorkspaceElements().forEach((workspaceElement) => {
                value.validateIfDeletedWorkspaceElement(workspaceElement);
                if (options.local && !options.remote) {
                    const localChange = {
                        state: workspaceElement.getState(),
                        fullName: workspaceElement.getFullName(),
                        type: workspaceElement.getMetadataName(),
                        filePath: workspaceElement.getSourcePath(),
                        deleteSupported: workspaceElement.getDeleteSupported()
                    };
                    this.localChanges.push(localChange);
                }
                else {
                    // if we want to find source conflicts between the workspace and the server,
                    // then pass along the locally changed workspace elements and
                    // populate this.localChanges during the _markConflicts() step
                    this.locallyChangedWorkspaceElements.push(workspaceElement);
                }
            });
        });
    }
    _populateServerChanges(options) {
        if (!options.remote) {
            return [];
        }
        return srcDevUtil.getMaxRevision(this.maxRevisionFile.path)
            .then((maxRevision) => this.force.toolingFind(this.scratchOrg, 'SourceMember', { RevisionNum: { $gt: maxRevision } }, ['MemberType', 'MemberName', 'IsNameObsolete']))
            .then((sourceMembers) => {
            sourceMembers.forEach(sourceMember => {
                const sourceMemberMetadataType = MetadataTypeFactory.getMetadataTypeFromMetadataName(sourceMember.MemberType, this.swa.metadataRegistry);
                if (sourceMemberMetadataType) {
                    sourceMember.MemberName = sourceMemberMetadataType.handleSlashesForSourceMemberName(sourceMember.MemberName);
                }
                if (this._forceIgnoreSourceMember(sourceMember, sourceMemberMetadataType)) {
                    return;
                }
                const remoteChangeElements = this._createRemoteChangeElementsFromSourceMember(sourceMember);
                this.remoteChanges = this.remoteChanges.concat(remoteChangeElements);
            });
        })
            .then(() => {
            this.remoteChanges = this.remoteChanges.filter(sm => !(sm.state === sourceState.DELETED && util.isNullOrUndefined(sm.filePath)));
        });
    }
    _forceIgnoreSourceMember(sourceMember, sourceMemberMetadataType) {
        // if user wants to ignore a permissionset with fullname abc then we check if forceignore denies abc.permissionset
        if (sourceMemberMetadataType) {
            const filename = `${sourceMemberMetadataType.getAggregateFullNameFromSourceMemberName(sourceMember.MemberName)}.${sourceMemberMetadataType.getExt()}`;
            if (this.forceIgnore.denies(filename)) {
                return true;
            }
        }
        return false;
    }
    _createRemoteChangeElementsFromSourceMember(sourceMember) {
        const remoteChangeElements = [];
        const sourceMemberMetadataType = MetadataTypeFactory.getMetadataTypeFromMetadataName(sourceMember.MemberType, this.swa.metadataRegistry);
        if (sourceMemberMetadataType) {
            if (sourceMemberMetadataType.trackRemoteChangeForSourceMemberName(sourceMember.MemberName)) {
                const correspondingWorkspaceElements = this._getCorrespondingWorkspaceElements(sourceMember, sourceMemberMetadataType);
                const state = this._getRemoteChangeState(sourceMember, correspondingWorkspaceElements);
                const metadataType = sourceMemberMetadataType.getMetadataName();
                if (this.swa.metadataRegistry.isSupported(metadataType)) {
                    if (!util.isNullOrUndefined(correspondingWorkspaceElements) && correspondingWorkspaceElements.length > 0 &&
                        !sourceMemberMetadataType.displayAggregateRemoteChangesOnly()) {
                        correspondingWorkspaceElements.forEach(workspaceElement => {
                            const remoteChangeElement = {
                                state,
                                fullName: workspaceElement.getFullName(),
                                type: sourceMemberMetadataType.getDisplayNameForRemoteChange(sourceMember.MemberType),
                                filePath: workspaceElement.getSourcePath()
                            };
                            remoteChangeElements.push(remoteChangeElement);
                        });
                    }
                    else {
                        const remoteChangeElement = {
                            state,
                            fullName: sourceMember.MemberName,
                            type: sourceMemberMetadataType.getDisplayNameForRemoteChange(sourceMember.MemberType)
                        };
                        remoteChangeElements.push(remoteChangeElement);
                    }
                }
            }
        }
        return remoteChangeElements;
    }
    _getRemoteChangeState(sourceMember, correspondingLocalWorkspaceElements) {
        if (sourceMember.IsNameObsolete) {
            return sourceState.DELETED;
        }
        else if (!util.isNullOrUndefined(correspondingLocalWorkspaceElements) && correspondingLocalWorkspaceElements.length > 0) {
            return sourceState.CHANGED;
        }
        else {
            return sourceState.NEW;
        }
    }
    _getCorrespondingWorkspaceElements(sourceMember, sourceMemberMetadataType) {
        const allLocalAggregateElements = this.swa.getAggregateSourceElements(false);
        if (!util.isNullOrUndefined(allLocalAggregateElements)) {
            if (sourceMemberMetadataType) {
                const aggregateFullName = sourceMemberMetadataType.getAggregateFullNameFromSourceMemberName(sourceMember.MemberName);
                const aggregateMetadataName = sourceMemberMetadataType.getAggregateMetadataName();
                const key = AggregateSourceElement.getKeyFromMetadataNameAndFullName(aggregateMetadataName, aggregateFullName);
                const localAggregateSourceElement = allLocalAggregateElements.get(key);
                if (!util.isNullOrUndefined(localAggregateSourceElement)) {
                    const workspaceElements = localAggregateSourceElement.getWorkspaceElements();
                    if (workspaceElements.length > 0) {
                        return workspaceElements.filter(workspaceElement => {
                            const workspaceElementMetadataType = MetadataTypeFactory.getMetadataTypeFromMetadataName(workspaceElement.getMetadataName(), this.swa.metadataRegistry);
                            return workspaceElementMetadataType.sourceMemberFullNameCorrespondsWithWorkspaceFullName(sourceMember.MemberName, workspaceElement.getFullName());
                        });
                    }
                }
            }
        }
        return [];
    }
    _markConflicts(options) {
        if (options.local && options.remote) {
            return this.locallyChangedWorkspaceElements.forEach(workspaceElement => {
                // a metadata element with same name and type modified both locally and in the server is considered a conflict
                const localChange = {
                    state: workspaceElement.getState(),
                    fullName: workspaceElement.getFullName(),
                    type: workspaceElement.getMetadataName(),
                    filePath: workspaceElement.getSourcePath(),
                    deleteSupported: workspaceElement.getDeleteSupported()
                };
                const workspaceElementMetadataType = MetadataTypeFactory.getMetadataTypeFromMetadataName(workspaceElement.getMetadataName(), this.swa.metadataRegistry);
                const remoteChanges = this.remoteChanges.filter(remoteChange => workspaceElementMetadataType.conflictDetected(remoteChange.type, remoteChange.fullName, workspaceElement.getFullName()));
                if (!util.isNullOrUndefined(remoteChanges) && remoteChanges.length > 0) {
                    localChange.isConflict = true;
                    remoteChanges.forEach(remoteChange => { remoteChange.isConflict = true; });
                }
                this.localChanges.push(localChange);
            });
        }
        else {
            return [];
        }
    }
    getLocalChanges() {
        return this.localChanges;
    }
    getRemoteChanges() {
        return this.remoteChanges;
    }
    getLocalConflicts() {
        const aggregateKeys = new Set();
        return this.localChanges.filter(item => {
            const metadataType = MetadataTypeFactory.getMetadataTypeFromMetadataName(item.type, this.swa.metadataRegistry);
            if (item.isConflict && metadataType.onlyDisplayOneConflictPerAggregate()) {
                const aggregateFullName = metadataType.getAggregateFullNameFromWorkspaceFullName(item.fullName);
                const key = `${metadataType.getMetadataName()}#${aggregateFullName}`;
                if (!aggregateKeys.has(key)) {
                    aggregateKeys.add(key);
                    return true;
                }
                return false;
            }
            return item.isConflict;
        }).map(item => {
            const metadataType = MetadataTypeFactory.getMetadataTypeFromMetadataName(item.type, this.swa.metadataRegistry);
            if (metadataType.onlyDisplayOneConflictPerAggregate()) {
                return {
                    state: item.state,
                    fullName: metadataType.getAggregateFullNameFromWorkspaceFullName(item.fullName),
                    type: item.type,
                    filePath: metadataType.getDisplayPathForLocalConflict(item.filePath),
                    deleteSupported: item.deleteSupported
                };
            }
            return item;
        });
    }
}
module.exports = SrcStatusApi;

//# sourceMappingURL=srcStatusApi.js.map
